import numpy as np
from .model import ForwardModel
from taurex.util.util import clip_native_to_wngrid


class SimpleForwardModel(ForwardModel):
    """ A 'simple' base model in the sense that its just
    a fairly standard single profiles model.
    It will handle settingup and initializing, building
    collecting parameters from given profiles for you.
    The only method that requires implementation is:

    - :func:`path_integral`

    Parameters
    ----------
    name: str
        Name to use in logging

    planet: :class:`~taurex.data.planet.Planet`, optional
        Planet model, default planet is Jupiter

    star: :class:`~taurex.data.stellar.star.Star`, optional
        Star model, default star is Sun-like

    pressure_profile: :class:`~taurex.data.profiles.pressure.pressureprofile.PressureProfile`, optional
        Pressure model, alternative is to set ``nlayers``, ``atm_min_pressure``
        and ``atm_max_pressure``

    temperature_profile: :class:`~taurex.data.profiles.temperature.tprofile.TemperatureProfile`, optional
        Temperature model, default is an :class:`~taurex.data.profiles.temperature.isothermal.Isothermal`
        profile at 1500 K

    chemistry: :class:`~taurex.data.profiles.chemistry.chemistry.Chemistry`, optional
        Chemistry model, default is
        :class:`~taurex.data.profiles.chemistry.taurexchemistry.TaurexChemistry` with
        ``H2O`` and ``CH4``

    nlayers: int, optional
        Number of layers. Used if ``pressure_profile`` is not defined.

    atm_min_pressure: float, optional
        Pressure at TOA. Used if ``pressure_profile`` is not defined.

    atm_max_pressure: float, optional
        Pressure at BOA. Used if ``pressure_profile`` is not defined.

    """

    def __init__(self, name,
                 planet=None,
                 star=None,
                 pressure_profile=None,
                 temperature_profile=None,
                 chemistry=None,
                 nlayers=100,
                 atm_min_pressure=1e-4,
                 atm_max_pressure=1e6):
        super().__init__(name)

        self._planet = planet
        self._star = star
        self._pressure_profile = pressure_profile
        self._temperature_profile = temperature_profile
        self._chemistry = chemistry
        self.debug('Passed: %s %s %s %s %s', planet, star, pressure_profile,
                   temperature_profile, chemistry)
        self.altitude_profile = None
        self.scaleheight_profile = None
        self.gravity_profile = None
        self._setup_defaults(nlayers, atm_min_pressure, atm_max_pressure)

        self._initialized = False

        self._sigma_opacities = None

        self._native_grid = None

    def _compute_inital_mu(self):
        from taurex.data.profiles.chemistry import TaurexChemistry, ConstantGas
        tc = TaurexChemistry()
        tc.addGas(ConstantGas('H2O'))
        self._inital_mu = tc

    def _setup_defaults(self, nlayers, atm_min_pressure, atm_max_pressure):

        if self._pressure_profile is None:
            from taurex.data.profiles.pressure import SimplePressureProfile
            self.info('No pressure profile defined, using simple pressure '
                      'profile with')
            self.info('parameters nlayers: %s, atm_pressure_range=(%s,%s)',
                      nlayers, atm_min_pressure, atm_max_pressure)

            self._pressure_profile = \
                SimplePressureProfile(nlayers, atm_min_pressure,
                                      atm_max_pressure)

        if self._planet is None:
            from taurex.data import Planet
            self.warning('No planet defined, using Jupiter as planet')
            self._planet = Planet()

        if self._temperature_profile is None:
            from taurex.data.profiles.temperature import Isothermal
            self.warning('No temeprature profile defined using default '
                         'Isothermal profile with T=1500 K')
            self._temperature_profile = Isothermal()

        if self._chemistry is None:
            from taurex.data.profiles.chemistry import TaurexChemistry, \
                ConstantGas
            tc = TaurexChemistry()
            self.warning('No gas profile set, using constant profile with H2O '
                         'and CH4')
            tc.addGas(ConstantGas('H2O', mix_ratio=1e-5))
            tc.addGas(ConstantGas('CH4', mix_ratio=1e-6))
            self._chemistry = tc

        if self._star is None:
            from taurex.data.stellar import BlackbodyStar
            self.warning('No star, using the Sun')
            self._star = BlackbodyStar()

    def initialize_profiles(self):
        """
        Initializes all profiles
        """

        self.info('Computing pressure profile')

        self.pressure.compute_pressure_profile()

        self._temperature_profile.initialize_profile(self._planet,
                                                     self.pressure.nLayers,
                                                     self.pressure.profile)

        # Initialize the atmosphere with a constant gas profile
        if self._initialized is False:
            self._inital_mu.initialize_chemistry(self.pressure.nLayers,
                                                 self.temperatureProfile,
                                                 self.pressureProfile,
                                                 None)

            self._compute_altitude_gravity_scaleheight_profile(
                self._inital_mu.muProfile)

            self._initialized = True

        # Now initialize the gas profile real
        self._chemistry.initialize_chemistry(self.pressure.nLayers,
                                             self.temperatureProfile,
                                             self.pressureProfile,
                                             self.altitude_profile)

        # Compute gravity scale height
        self._compute_altitude_gravity_scaleheight_profile()

    def collect_fitting_parameters(self):
        """
        Collects all fitting parameters from all
        profiles within the forward model
        """

        self._fitting_parameters = {}
        self._fitting_parameters.update(self.fitting_parameters())
        self._fitting_parameters.update(self._planet.fitting_parameters())
        if self._star is not None:
            self._fitting_parameters.update(self._star.fitting_parameters())
        self._fitting_parameters.update(self.pressure.fitting_parameters())

        self._fitting_parameters.update(
            self._temperature_profile.fitting_parameters())

        self._fitting_parameters.update(self._chemistry.fitting_parameters())

        for contrib in self.contribution_list:
            self._fitting_parameters.update(contrib.fitting_parameters())

        self.debug('Available Fitting params: %s',
                   list(self._fitting_parameters.keys()))

    def build(self):
        """
        Build the forward model. Must be called at least
        once before running :func:`model`
        """

        self.contribution_list.sort(key=lambda x: x.order)

        self.info('Building model........')
        self._compute_inital_mu()
        self.info('Collecting paramters')
        self.collect_fitting_parameters()
        self.info('Setting up profiles')
        self.initialize_profiles()

        self.info('Setting up contributions')
        for contrib in self.contribution_list:
            contrib.build(self)
        self.info('DONE')

    # altitude, gravity and scale height profile
    def _compute_altitude_gravity_scaleheight_profile(self, mu_profile=None):
        """
        Computes altitude, gravity and scale height of the atmosphere.
        Only call after :func:`build` has been called at least once.

        Parameters
        ----------
        mu_profile, optional:
            Molecular weight profile at each layer

        """

        from taurex.constants import KBOLTZ
        if mu_profile is None:
            mu_profile = self._chemistry.muProfile

        # build the altitude profile from the bottom up
        nlayers = self.pressure.nLayers
        H = np.zeros(nlayers)
        g = np.zeros(nlayers)
        z = np.zeros(nlayers)

        # surface gravity (0th layer)
        g[0] = self._planet.gravity
        # scaleheight at the surface (0th layer)
        H[0] = (KBOLTZ*self.temperatureProfile[0])/(mu_profile[0]*g[0])

        for i in range(1, nlayers):
            deltaz = (-1.)*H[i-1]*np.log(
                self.pressure.pressure_profile_levels[i] /
                self.pressure.pressure_profile_levels[i-1])

            z[i] = z[i-1] + deltaz  # altitude at the i-th layer

            with np.errstate(over='ignore'):
                # gravity at the i-th layer
                g[i] = self._planet.gravity_at_height(z[i])
                self.debug('G[%s] = %s', i, g[i])

            with np.errstate(divide='ignore'):
                H[i] = (KBOLTZ*self.temperatureProfile[i])/(mu_profile[i]*g[i])

        self.altitude_profile = z
        self.scaleheight_profile = H
        self.gravity_profile = g

    @property
    def pressureProfile(self):
        """
        Atmospheric pressure profile in Pa
        """
        return self.pressure.profile

    @property
    def temperatureProfile(self):
        """
        Atmospheric temperature profile in K
        """
        return self._temperature_profile.profile

    @property
    def densityProfile(self):
        """
        Atmospheric density profile in m-3
        """
        from taurex.constants import KBOLTZ
        return (self.pressureProfile)/(KBOLTZ*self.temperatureProfile)

    @property
    def altitudeProfile(self):
        """
        Atmospheric altitude profile in m
        """
        return self.altitude_profile

    @property
    def nLayers(self):
        """
        Number of layers
        """
        return self.pressure.nLayers

    @property
    def chemistry(self):
        """
        Chemistry model
        """
        return self._chemistry

    @property
    def pressure(self):
        """
        Pressure model
        """
        return self._pressure_profile

    @property
    def temperature(self):
        """
        Temperature model
        """
        return self._temperature_profile

    @property
    def star(self):
        """
        Star model
        """
        return self._star

    @property
    def planet(self):
        """
        Planet model
        """
        return self._planet

    @property
    def nativeWavenumberGrid(self):
        """

        Searches through active molecules to determine the
        native wavenumber grid

        Returns
        -------

        wngrid: :obj:`array`
            Native grid

        Raises
        ------
        InvalidModelException
            If no active molecules in atmosphere
        """
        from taurex.exceptions import InvalidModelException
        from taurex.cache.opacitycache import OpacityCache

        active_gases = self.chemistry.activeGases

        wavenumbergrid = \
            [OpacityCache()[gas].wavenumberGrid for gas in active_gases]

        current_grid = None
        for wn in wavenumbergrid:
            if current_grid is None:
                current_grid = wn
            if wn.shape[0] > current_grid.shape[0]:
                current_grid = wn

        if current_grid is None:
            self.error('No active molecules detected')
            self.error('Most likely no cross-sections were detected')
            raise InvalidModelException('No active absorbing molecules')

        return current_grid

    def model(self, wngrid=None, cutoff_grid=True):
        """
        Runs the forward model

        Parameters
        ----------

        wngrid: :obj:`array`, optional
            Wavenumber grid, default is to use native grid

        cutoff_grid: bool
            Run model only on ``wngrid`` given, default is ``True``

        Returns
        -------

        native_grid: :obj:`array`
            Native wavenumber grid, clipped if ``wngrid`` passed

        depth: :obj:`array`
            Resulting depth

        tau: :obj:`array`
            Optical depth.

        extra: ``None``
            Empty
        """

        # Setup profiles
        self.initialize_profiles()

        # Clip grid if necessary
        native_grid = self.nativeWavenumberGrid
        if wngrid is not None and cutoff_grid:
            native_grid = clip_native_to_wngrid(native_grid, wngrid)

        # Initialize star
        self._star.initialize(native_grid)

        # Prepare contributions
        for contrib in self.contribution_list:
            contrib.prepare(self, native_grid)

        # Compute path integral
        absorp, tau = self.path_integral(native_grid, False)

        return native_grid, absorp, tau, None

    def model_contrib(self, wngrid=None, cutoff_grid=True):
        """
        Models each contribution seperately
        """
        # Setup profiles
        self.initialize_profiles()

        # Copy over contribution list
        full_contrib_list = self.contribution_list
        # Get the native grid
        native_grid = self.nativeWavenumberGrid

        # Clip grid
        all_contrib_dict = {}
        if wngrid is not None and cutoff_grid:
            native_grid = clip_native_to_wngrid(native_grid, wngrid)

        # Initialize star
        self._star.initialize(native_grid)

        for contrib in full_contrib_list:
            self.contribution_list = [contrib]
            contrib.prepare(self, native_grid)
            absorp, tau = self.path_integral(native_grid, False)
            all_contrib_dict[contrib.name] = (absorp, tau, None)

        self.contribution_list = full_contrib_list
        return native_grid, all_contrib_dict

    def model_full_contrib(self, wngrid=None, cutoff_grid=True):
        """

        Like :func:`model_contrib` except all components for
        each contribution are modelled

        """
        native_grid = self.nativeWavenumberGrid
        if wngrid is not None and cutoff_grid:
            native_grid = clip_native_to_wngrid(native_grid, wngrid)

        self.initialize_profiles()
        self._star.initialize(native_grid)

        result_dict = {}

        full_contrib_list = self.contribution_list

        self.debug('NATIVE GRID %s', native_grid.shape)

        self.info('Modelling each contribution.....')
        for contrib in full_contrib_list:
            self.contribution_list = [contrib]
            contrib_name = contrib.name
            contrib_res_list = []

            for name, __ in contrib.prepare_each(self, native_grid):
                self.info('\t%s---%s contribtuion', contrib_name, name)
                absorp, tau = self.path_integral(native_grid, False)
                contrib_res_list.append((name, absorp, tau, None))

            result_dict[contrib_name] = contrib_res_list

        self.contribution_list = full_contrib_list
        return native_grid, result_dict

    def compute_error(self, samples, wngrid=None, binner=None):
        """

        Computes standard deviations from samples

        Parameters
        ----------

        samples: 

        """
        from taurex.util.math import OnlineVariance
        tp_profiles = OnlineVariance()
        active_gases = OnlineVariance()
        inactive_gases = OnlineVariance()

        if binner is not None:
            binned_spectrum = OnlineVariance()
        else:
            binned_spectrum = None
        native_spectrum = OnlineVariance()

        for weight in samples():

            native_grid, native, tau, _ = self.model(wngrid=wngrid,
                                                     cutoff_grid=False)

            tp_profiles.update(self.temperatureProfile, weight=weight)
            active_gases.update(self.chemistry.activeGasMixProfile,
                                weight=weight)
            inactive_gases.update(self.chemistry.inactiveGasMixProfile,
                                  weight=weight)

            native_spectrum.update(native, weight=weight)

            if binned_spectrum is not None:
                binned = binner.bindown(native_grid, native)[1]
                binned_spectrum.update(binned, weight=weight)

        profile_dict = {}
        spectrum_dict = {}

        tp_std = np.sqrt(tp_profiles.parallelVariance())
        active_std = np.sqrt(active_gases.parallelVariance())
        inactive_std = np.sqrt(inactive_gases.parallelVariance())

        profile_dict['temp_profile_std'] = tp_std
        profile_dict['active_mix_profile_std'] = active_std
        profile_dict['inactive_mix_profile_std'] = inactive_std

        spectrum_dict['native_std'] = \
            np.sqrt(native_spectrum.parallelVariance())

        if binned_spectrum is not None:
            spectrum_dict['binned_std'] = \
                np.sqrt(binned_spectrum.parallelVariance())

        return profile_dict, spectrum_dict

    def path_integral(self, wngrid, return_contrib):
        raise NotImplementedError

    def write(self, output):
        # Run a model if needed
        self.model()

        model = super().write(output)

        self._chemistry.write(model)
        self._temperature_profile.write(model)
        self.pressure.write(model)
        self._planet.write(model)
        self._star.write(model)
        return model



from .emission import EmissionModel
from taurex.constants import PI


class DirectImageModel(EmissionModel):
    """
    A forward model for direct imaging of exo-planets

    Parameters
    ----------

    planet: :class:`~taurex.data.planet.Planet`, optional
        Planet model, default planet is Jupiter

    star: :class:`~taurex.data.stellar.star.Star`, optional
        Star model, default star is Sun-like

    pressure_profile: :class:`~taurex.data.profiles.pressure.pressureprofile.PressureProfile`, optional
        Pressure model, alternative is to set ``nlayers``, ``atm_min_pressure``
        and ``atm_max_pressure``

    temperature_profile: :class:`~taurex.data.profiles.temperature.tprofile.TemperatureProfile`, optional
        Temperature model, default is an :class:`~taurex.data.profiles.temperature.isothermal.Isothermal`
        profile at 1500 K

    chemistry: :class:`~taurex.data.profiles.chemistry.chemistry.Chemistry`, optional
        Chemistry model, default is
        :class:`~taurex.data.profiles.chemistry.taurexchemistry.TaurexChemistry` with
        ``H2O`` and ``CH4``

    nlayers: int, optional
        Number of layers. Used if ``pressure_profile`` is not defined.

    atm_min_pressure: float, optional
        Pressure at TOA. Used if ``pressure_profile`` is not defined.

    atm_max_pressure: float, optional
        Pressure at BOA. Used if ``pressure_profile`` is not defined.

    ngauss: int, optional
        Number of gaussian quadrature points, default = 4

    """

    def __init__(self,
                 planet=None,
                 star=None,
                 pressure_profile=None,
                 temperature_profile=None,
                 chemistry=None,
                 nlayers=100,
                 atm_min_pressure=1e-4,
                 atm_max_pressure=1e6,
                 ngauss=4,
                 ):
        super().__init__(planet,
                         star,
                         pressure_profile,
                         temperature_profile,
                         chemistry,
                         nlayers,
                         atm_min_pressure,
                         atm_max_pressure,
                         ngauss=ngauss)

    def compute_final_flux(self, f_total):
        star_distance_meters = self._star.distance*3.08567758e16

        SDR = pow((star_distance_meters/3.08567758e16), 2)
        SDR = 1.0
        planet_radius = self._planet.fullRadius

        return((f_total * (planet_radius**2) * 2.0 * PI) /
               (4 * PI * (star_distance_meters**2))) * SDR



"""
Contains caching class for Molecular cross section files
"""

from .singleton import Singleton
from taurex.log import Logger
import pathlib
class OpacityCache(Singleton):
    """
    Implements a lazy load of opacities. A singleton that
    loads and caches xsections as they are needed. Calling

    >>> opt = OpacityCache()
    >>> opt2 = OpacityCache()

    Reveals that:

    >>> opt == opt2
    True
    
    Importantly this class will automatically search directories for cross-sections
    set using the :func:`set_opacity_path` method:

    >>> opt.set_opacity_path('path/to/crossections')

    Multiple paths can be set as well

    >>> opt.set_opacity_path(['/path/to/crosssections','/another/path/to/crosssections'])

    Currently only :obj:`.pickle` files are supported.

    To get the cross-section object for a particular molecule use the square bracket operator:

    >>> opt['H2O']
    <taurex.opacity.pickleopacity.PickleOpacity at 0x107a60be0>

    This returns a :class:`~taurex.opacity.pickleopacity.PickleOpacity` object for you to compute H2O cross sections from.
    When called for the first time, a directory search is performed and, if found, the appropriate cross-section is loaded. Subsequent calls will immediately
    return the already loaded object:

    >>> h2o_a = opt['H2O']
    >>> h2o_b = opt['H2O']
    >>> h2o_a == h2o_b
    True

    Lastly if you've got a hot new opacity format, you can try out
    by manually adding it into the cache:

    >>> new_h2o = MyNewOpacityFormat()
    >>> opt.add_opacity(new_h2o)
    >>> opt['H2O]
    <MyNewOpacityFormat at 0x107a60be0>

    Now TauREx3 will use it instead in all calculations!

    """
    def init(self):
        self.opacity_dict = {}
        self._opacity_path = None
        self.log = Logger('OpacityCache')
        self._default_interpolation = 'linear'
        self._memory_mode = True
        self._radis = False
        self._radis_props = (600, 30000, 10000)
    
    def set_opacity_path(self, opacity_path):
        """
        Set the path(s) that will be searched for opacities.
        Opacities in this path must be of supported types:

            - HDF5 opacities
            - ``.pickle`` opacities
            - ExoTransmit opacities.

        Parameters
        ----------

        opacity_path : str or :obj:`list` of str, optional
            search path(s) to look for molecular opacities

        """

        import os
        if not os.path.isdir(opacity_path):
            self.log.error('PATH: %s does not exist!!!', opacity_path)
            raise NotADirectoryError
        self.log.debug('Path set to %s', opacity_path)
        self._opacity_path = opacity_path

    def enable_radis(self, enable):
        """
        Enables/Disables use of RADIS to fill in missing molecules
        using HITRAN.

        .. warning::
            This is extremely unstable and crashes frequently.
            It is also very slow as it requires
            the computation of the Voigt profile for every temperature.
            We recommend leaving it as False unless necessary.

        Parameters
        ----------
        enable: bool
            Whether to enable RADIS functionality (default = False)

        """

        self._radis = enable

    def set_radis_wavenumber(self, wn_start, wn_end, wn_points):
        self._radis_props = (wn_start, wn_end, wn_points)

        

        self.clear_cache()
    
    def set_memory_mode(self,in_memory):
        """
        If using the HDF5 opacities, whether to stream
        opacities from file (slower, less memory) or load
        them into memory (faster, more memory)

        Parameters
        ----------
        in_memory: bool
            Whether HDF5 files should be streamed (False)
            or loaded into memory (True, default)


        """

        self._memory_mode = in_memory
        self.clear_cache()

    def set_interpolation(self,interpolation_mode):
        """
        Sets the interpolation mode for all currently loaded (and future loaded) cross-sections

        Can either be ``linear`` for linear interpolation of both temeprature and pressure:

        >>> OpacityCache().set_interpolation('linear')

        or ``exp`` for natural exponential interpolation of temperature
        and linear for pressure

        >>> OpacityCache().set_interpolation('exp')
        

        Parameters
        ----------
        interpolation_mode: str
            Either ``linear`` for bilinear interpolation or ``exp`` for exp-linear interpolation
        

        """
        return
        self._default_interpolation = interpolation_mode
        for values in self.opacity_dict.values():
            values.set_interpolation_mode(self._default_interpolation)
    
    

    def __getitem__(self,key):
        """
        For a molecule return the relevant :class:`~taurex.opacity.opacity.Opacity` object.


        Parameter
        ---------
        key : str
            molecule name

        Returns
        -------
        :class:`~taurex.opacity.pickleopacity.PickleOpacity`
            Cross-section object desired
        
        Raise
        -----
        Exception
            If molecule could not be loaded/found

        """
        if key in self.opacity_dict:
            return self.opacity_dict[key]
        else:
            #Try a load of the opacity
            self.load_opacity(molecule_filter=[key])
            #If we have it after a load then good job boys
            if key in self.opacity_dict:
                return self.opacity_dict[key]
            else:
                try:
                    if self._radis:
                        return self.create_radis_opacity(key,molecule_filter=[key])
                    else:
                        raise Exception
                except Exception as e:
                    self.log.error('EXception thrown %s',e)
                    #Otherwise throw an error
                    self.log.error('Opacity for molecule %s could not be loaded',key)
                    self.log.error('It could not be found in the local dictionary %s',list(self.opacity_dict.keys()))
                    self.log.error('Or paths %s',self._opacity_path)
                    self.log.error('Try loading it manually/ putting it in a path')
                    raise Exception('Opacity could notn be loaded')

    def create_radis_opacity(self,molecule,molecule_filter=None):
        from taurex.opacity.radisopacity import RadisHITRANOpacity
        if molecule not in self.opacity_dict:
            self.log.info('Creating Opacity from RADIS+HITRAN')
            wn_start,wn_end,wn_points = self._radis_props
            radis = RadisHITRANOpacity(molecule_name=molecule, wn_start=wn_start,wn_end=wn_end,wn_points=wn_points)
            
            

            self.add_opacity(radis,molecule_filter=molecule_filter)
            return radis
        else:
            self.log.info('Opacity %s already exsits',molecule)


    def add_opacity(self,opacity,molecule_filter=None):
        """

        Adds a :class:`~taurex.opacity.opacity.Opacity` object to the cache to then be
        used by Taurex 3

        Parameters
        ----------
        opacity : :class:`~taurex.opacity.opacity.Opacity`
            Opacity object to add to the cache
        
        molecule_filter : :obj:`list` of str , optional
            If provided, the opacity object will only be included
            if its molecule is in the list. Mostly used by the 
            :func:`__getitem__` for filtering

        """
        self.log.info('Reading opacity %s',opacity.moleculeName)
        if opacity.moleculeName in self.opacity_dict:
            self.log.warning('Opacity with name %s already in opactiy dictionary %s skipping',opacity.moleculeName,self.opacity_dict.keys())
            return
        if molecule_filter is not None:
            if opacity.moleculeName in molecule_filter:
                self.log.info('Loading opacity %s into model',opacity.moleculeName)
                self.opacity_dict[opacity.moleculeName] = opacity       
        else:     
            self.log.info('Loading opacity %s into model',opacity.moleculeName)
            self.opacity_dict[opacity.moleculeName] = opacity   


    def search_hdf5_molecules(self):
        """
        Find molecules with HDF5 opacities in set path

        Returns
        -------
        molecules: :obj`list`
            List of molecules with HDF5 opacities
        
        """
        from glob import glob
        import os    
        from taurex.opacity.hdf5opacity import HDF5Opacity
        glob_path = [os.path.join(self._opacity_path,'*.h5'),os.path.join(self._opacity_path,'*.hdf5')]
        file_list = [f for glist in glob_path for f in glob(glist)]
        
        return [HDF5Opacity(f,interpolation_mode=self._default_interpolation,in_memory=False).moleculeName for f in file_list ]

    def search_pickle_molecules(self):
        """
        Find molecules with ``.pickle`` opacities in set path

        Returns
        -------
        molecules: :obj`list`
            List of molecules with ``.pickle`` opacities
        
        """

        from glob import glob
        import os    
        glob_path = os.path.join(self._opacity_path,'*.pickle')
        file_list = [f for f in glob(glob_path)]
        
        return [pathlib.Path(f).stem.split('.')[0] for f in file_list ]

    def search_exotransmit_molecules(self):
        """
        Find molecules with Exo-Transmit opacities in set path

        Returns
        -------
        molecules: :obj`list`
            List of molecules with ExoTransmit opacities
        
        """

        from glob import glob
        import os    
        glob_path = os.path.join(self._opacity_path,'*.dat')
        file_list = [f for f in glob(glob_path)]
        
        return [pathlib.Path(f).stem[4:] for f in file_list ]

    def search_radis_molecules(self):
        """
        Searches for molecules in HITRAN

        Returns
        -------
        molecules: :obj`list`
            List of molecules available in HITRAN, if radis is enabled,
            otherwise an empty list

        """
        trans = { '1':'H2O',    '2':'CO2',   '3':'O3',      '4':'N2O',   '5':'CO',    '6':'CH4',   '7':'O2',     
            '9':'SO2',   '10':'NO2',  '11':'NH3',    '12':'HNO3', '13':'OH',   '14':'HF',   '15':'HCl',   '16':'HBr',
            '17':'HI',    '18':'ClO',  '19':'OCS',    '20':'H2CO', '21':'HOCl',    '23':'HCN',   '24':'CH3Cl',
            '25':'H2O2',  '26':'C2H2', '27':'C2H6',   '28':'PH3',  '29':'COF2', '30':'SF6',  '31':'H2S',   '32':'HCOOH',
            '33':'HO2',   '34':'O',    '35':'ClONO2', '36':'NO+',  '37':'HOBr', '38':'C2H4',  '40':'CH3Br',
            '41':'CH3CN', '42':'CF4',  '43':'C4H2',   '44':'HC3N',   '46':'CS',   '47':'SO3'}
        if self._radis:
            return list(trans.values())
        else:
            return []
    def find_list_of_molecules(self):
        from glob import glob
        import os
        from taurex.opacity import PickleOpacity
        pickles = []
        hedef = []
        exo = []
        if self._opacity_path is not None:
        
            pickles = self.search_pickle_molecules()
            hedef = self.search_hdf5_molecules()
            exo = self.search_exotransmit_molecules()
        return list(set(pickles+hedef+exo+self.search_radis_molecules()))
    def load_opacity_from_path(self,path,molecule_filter=None):
        """
        Searches path for molecular cross-section files, creates and loads them into the cache
        ``.pickle`` will be loaded as :class:`~taurex.opacity.pickleopacity.PickleOpacity`
        

        Parameters
        ----------
        path : str
            Path to search for molecular cross-section files
        
        molecule_filter : :obj:`list` of str , optional
            If provided, the opacity will only be loaded
            if its molecule is in this list. Mostly used by the 
            :func:`__getitem__` for filtering

        """ 
        from glob import glob
        import os
        from taurex.opacity import PickleOpacity
        from taurex.opacity.hdf5opacity import HDF5Opacity
        from taurex.opacity.exotransmit import ExoTransmitOpacity
        glob_path = [os.path.join(path,'*.h5'),os.path.join(path,'*.hdf5'),os.path.join(path,'*.pickle'),os.path.join(path,'*.dat')]
    
        file_list = [f for glist in glob_path for f in glob(glist)]
        self.log.debug('File list %s',file_list)
        for files in file_list:
            op = None
            if files.lower().endswith(('.hdf5', '.h5')):
                op = HDF5Opacity(files,interpolation_mode=self._default_interpolation,in_memory=False)
                
                if molecule_filter is not None:
                        if not op.moleculeName in molecule_filter:
                            continue
                if op.moleculeName in self.opacity_dict.keys():
                    continue
                del op

                op = HDF5Opacity(files,interpolation_mode=self._default_interpolation,in_memory=self._memory_mode)

            elif files.endswith('pickle'):
                splits = pathlib.Path(files).stem.split('.')
                if molecule_filter is not None:
                        if not splits[0] in molecule_filter:
                            continue
                if splits[0] in self.opacity_dict.keys():
                    continue
                op = PickleOpacity(files,interpolation_mode=self._default_interpolation)
                op._molecule_name = splits[0]
            elif files.endswith('dat'):
                mol_name = pathlib.Path(files).stem[4:]
                if molecule_filter is not None:
                        if not mol_name in molecule_filter:
                            continue
                if mol_name in self.opacity_dict.keys():
                    continue
                op = ExoTransmitOpacity(files,interpolation_mode=self._default_interpolation)
            if op is not None:
                self.add_opacity(op,molecule_filter=molecule_filter)

    def load_opacity(self,opacities=None,opacity_path=None,molecule_filter=None):
        """
        Main function to use when loading molecular opacities. Handles both 
        cross sections and paths. Handles lists of either so lists of 
        :class:`~taurex.opacity.opacity.Opacity` objects or lists of paths can be used
        to load multiple files/objects


        Parameters
        ----------
        opacities : :class:`~taurex.opacity.opacity.Opacity` or :obj:`list` of :class:`~taurex.opacity.opacity.Opacity` , optional
            Object(s) to include in cache
        
        opacity_path : str or :obj:`list` of str, optional
            search path(s) to look for molecular opacities
        
        molecule_filter : :obj:`list` of str , optional
            If provided, the opacity will only be loaded
            if its molecule is in this list. Mostly used by the 
            :func:`__getitem__` for filtering

        """ 
        from taurex.opacity import Opacity
        
        if opacity_path is None:
            opacity_path = self._opacity_path

        if opacities is not None:
            if isinstance(opacities,(list,)):
                self.log.debug('Opacity passed is list')
                for opacity in opacities:
                    self.add_opacity(opacity,molecule_filter=molecule_filter)
            elif isinstance(opacities,Opacity):
                self.add_opacity(opacities,molecule_filter=molecule_filter)
            else:
                self.log.error('Unknown type %s passed into opacities, should be a list, single \
                     opacity or None if reading a path',type(opacities))
                raise Exception('Unknown type passed into opacities')
        elif opacity_path is not None:

            if isinstance(opacity_path,str):
                self.load_opacity_from_path(opacity_path,molecule_filter=molecule_filter)
            elif isinstance(opacity_path,(list,)):
                for path in opacity_path:
                    self.load_opacity_from_path(path,molecule_filter=molecule_filter)
    
    def clear_cache(self):
        """
        Clears all currently loaded cross-sections
        """
        self.opacity_dict = {}


from .binner import Binner
from taurex.util.util import compute_bin_edges
import numpy as np
from taurex import OutputSize


class FluxBinner(Binner):
    """
    Bins to a wavenumber grid given by ``wngrid`` using a
    more accurate method that takes into account the amount
    of contribution from each native bin. This method also
    handles cases where bins are not continuous and/or
    overlapping.

    Parameters
    ----------

    wngrid: :obj:`array`
        Wavenumber grid

    wngrid_width: :obj:`array`, optional
        Must have same shape as ``wngrid``
        Full bin widths for each wavenumber grid point
        given in ``wngrid``. If not provided then
        this is automatically computed from ``wngrid``.

    """
    def __init__(self, wngrid, wngrid_width=None):
        super().__init__()

        sort_grid = wngrid.argsort()
        self._wngrid = wngrid[sort_grid]
        self._wngrid_width = wngrid_width

        if self._wngrid_width is None:
            self._wngrid_width = compute_bin_edges(self._wngrid)[-1]
        elif hasattr(self._wngrid_width, '__len__'):
            if len(self._wngrid_width) != len(self._wngrid):
                raise ValueError('Wavenumber width should be signel value or '
                                 'same shape as wavenumber grid')
            self._wngrid_width = wngrid_width[sort_grid]

        if not hasattr(self._wngrid_width, '__len__'):
            self._wngrid_width = np.ones_like(self._wngrid)*self._wngrid_width

    def bindown(self, wngrid, spectrum, grid_width=None, error=None):
        """

        Bins down spectrum.

        Parameters
        ----------
        wngrid : :obj:`array`
            The wavenumber grid of the spectrum to be binned down.

        spectrum: :obj:`array`
            The spectra we wish to bin-down. Must be same shape as
            ``wngrid``.

        grid_width: :obj:`array`, optional
            Wavenumber grid full-widths for the spectrum to be binned down.
            Must be same shape as ``wngrid``.
            Optional.

        error: :obj:`array`, optional
            Associated errors or noise of the spectrum. Must be same shape
            as ``wngrid``.Optional parameter.

        Returns
        -------
        binned_wngrid : :obj:`array`
            New wavenumber grid

        spectrum: :obj:`array`
            Binned spectrum.

        grid_width: :obj:`array`
            New grid-widths

        error: :obj:`array` or None
            Binned error if given else ``None``

        """

        sorted_input = wngrid.argsort()
        spectrum = spectrum[..., sorted_input]
        if error is not None:
            error = error[..., sorted_input]

        bin_spectrum = np.zeros(spectrum[..., 0].shape + self._wngrid.shape)

        if error is not None:
            bin_error = np.zeros(spectrum[..., 0].shape + self._wngrid.shape)
        else:
            bin_error = None

        old_spect_wn = wngrid

        old_spect_flux = spectrum
        old_spect_err = error

        old_spect_width = grid_width

        if old_spect_width is None:
            old_spect_width = compute_bin_edges(old_spect_wn)[-1]

        old_spect_min = old_spect_wn - old_spect_width/2
        old_spect_max = old_spect_wn + old_spect_width/2

        new_spec_lhs = self._wngrid
        new_spec_rhs = self._wngrid_width

        new_spec_wn = self._wngrid

        new_spec_wn_min = new_spec_lhs - new_spec_rhs/2
        new_spec_wn_max = new_spec_lhs + new_spec_rhs/2

        save_start = 0
        save_stop = 0

        for idx, res in enumerate(zip(new_spec_wn, new_spec_wn_min,
                                      new_spec_wn_max)):

            wn, wn_min, wn_max = res
            sum_spectrum = 0
            sum_noise = 0
            sum_weight = 0

            save_start = np.searchsorted(old_spect_max, wn_min, side='right')
            save_stop = np.searchsorted(old_spect_min[1:], wn_max,
                                        side='right')

            save_stop = min(save_stop, old_spect_min.shape[0]-1)
            save_start = min(save_start, old_spect_min.shape[0]-1)

            if not wn_min <= old_spect_max[save_start] or not \
                    old_spect_min[save_stop] <= wn_max:
                continue

            spect_min = old_spect_min[save_start:save_stop+1]
            spect_max = old_spect_max[save_start:save_stop+1]

            weight = (np.minimum(wn_max, spect_max) -
                      np.maximum(spect_min, wn_min))/(wn_max-wn_min)

            sum_weight = np.sum(weight)

            sum_spectrum = np.sum(weight *
                                  old_spect_flux[..., save_start:save_stop+1],
                                  axis=-1)

            if error is not None:
                sum_noise = np.sum(weight * weight *
                                   old_spect_err[..., save_start:save_stop+1]**2,
                                   axis=0)

                sum_noise = np.sqrt(sum_noise / sum_weight/sum_weight)

            bin_spectrum[..., idx] = sum_spectrum

            if error is not None:
                bin_error[idx] = sum_noise

        return self._wngrid, bin_spectrum, bin_error, self._wngrid_width

    def generate_spectrum_output(self, model_output,
                                 output_size=OutputSize.heavy):

        output = super().generate_spectrum_output(model_output,
                                                  output_size=output_size)
        output['binned_wngrid'] = self._wngrid
        output['binned_wlgrid'] = 10000/self._wngrid
        output['binned_wnwidth'] = self._wngrid_width
        output['binned_wlwidth'] = 1.0/self._wngrid_width
        return output



from .data.profiles.temperature import Isothermal
from .data.profiles.temperature import Guillot2010
from .data.profiles.temperature import Rodgers2000
from .data.profiles.temperature import TemperatureProfile
from .data.profiles.temperature import TemperatureFile


"""Optimized Math functions used in taurex"""

import numba
import numpy as np

from numba import vectorize, float64
import math

@numba.vectorize([float64(float64,float64,float64, float64)])
def _expstage0(x1,x2, x11, x21):
    return x1*x11 - x2*(x11-x21)
@numba.vectorize([float64(float64,float64)],fastmath=True)
def _expstage1(x1, x2):
    return math.log(x1/x2)

@numba.vectorize([float64(float64,float64)],fastmath=True)
def _expstage2(C,x):
    return C*x

@numba.vectorize([float64(float64,float64,float64)],fastmath=True)
def _expstage3(C,x1,x2):
    return C*x1*x2

@numba.njit(nogil=True,fastmath=True)
def interp_exp_and_lin_broken(x11, x12, x21, x22, T, Tmin, Tmax, P, Pmin, Pmax):
    res = np.zeros_like(x11)
    x0 = -Pmin
    x1 = Pmax + x0
    x2 = P + x0
    factor1 = 1.0/(T*(Tmax - Tmin))
    factor2 = 1.0/x1
    x3 = _expstage0(x1,x2,x11,x21)
    x4 = _expstage0(x1,x2,x12,x22)
    x5 = _expstage1(x3,x4)
    x6 = _expstage2(Tmax*(-T + Tmin)*factor1,x5)
    x7 = _expstage3(factor2,x3,x6)
    for i in range(x11.shape[0]):
        res[i] = x7[i]*math.exp(x6[i])
    return res 


def interp_exp_and_lin(x11, x12, x21, x22, T, Tmin, Tmax, P, Pmin, Pmax):
        """
        2D interpolation

        Applies linear interpolation across P and natural exp interpolation
        across T between Pmin,Pmax and Tmin,Tmax

        Parameters
        ----------
        x11: array
            Array corresponding to Pmin,Tmin
        
        x12: array
            Array corresponding to Pmin,Tmax
        
        x21: array 
            Array corresponding to Pmax,Tmin
        
        x22: array
            Array corresponding to Pmax,Tmax

        T: float
            Coordinate to exp interpolate to

        Tmin: float
            Nearest known T coordinate where Tmin < T
        
        Tmax: float
            Nearest known T coordinate where T < Tmax

        P: float
            Coordinate to linear interpolate to

        Pmin: float
            Nearest known P coordinate where Pmin < P
        
        Pmax: float
            Nearest known P coordinate where P < Tmax
        
        """

        return ne.evaluate('((x11*(Pmax - Pmin) - (P - Pmin)*(x11 - x21))*exp(Tmax*(-T + Tmin)*log((x11*(Pmax - Pmin) - (P - Pmin)*(x11 - x21))/(x12*(Pmax - Pmin) - (P - Pmin)*(x12 - x22)))/(T*(Tmax - Tmin)))/(Pmax - Pmin))')


def interp_exp_only(x11,x12,T,Tmin,Tmax):
    return x11*np.exp(Tmax*(-T + Tmin)*np.log(x11/x12)/(T*(Tmax - Tmin)))

def interp_lin_only(x11,x12,P,Pmin,Pmax):
    return (x11*(Pmax - Pmin) - (P - Pmin)*(x11 - x12))/(Pmax - Pmin)


def intepr_bilin_old(x11, x12, x21, x22, T, Tmin, Tmax, P, Pmin, Pmax):
    return (x11*(Pmax - Pmin)*(Tmax - Tmin) - (P - Pmin)*(Tmax - Tmin)*(x11 - x21) - (T - Tmin)*(-(P - Pmin)*(x11 - x21) + (P - Pmin)*(x12 - x22) + (Pmax - Pmin)*(x11 - x12)))/((Pmax - Pmin)*(Tmax - Tmin))


def compute_rayleigh_cross_section(wngrid,n,n_air = 2.6867805e25,king=1.0):
    wlgrid = (10000/wngrid)*1e-6

    
    n_factor = (n**2 - 1)/(n_air*(n**2 + 2))
    sigma = 24.0*(np.pi**3)*king*(n_factor**2)/(wlgrid**4)

    return sigma

def test_nan(val):
    if hasattr(val,'__len__'):
        try:
            return np.isnan(val).any()
        except TypeError:
           # print(type(val))
            return True
    else:
        return val != val


@numba.vectorize([float64(float64,float64,float64)],fastmath=True)
def _linstage0(x11,x21,x):
    return x*(x11-x21)

@numba.njit(nogil=True,fastmath=True)
def intepr_bilin(x11, x12, x21, x22, T, Tmin, Tmax, P, Pmin, Pmax):
    x0 = -Pmin    
    x1 = Pmax + x0
    x2 = -Tmin    
    x3 = Tmax + x2
    x4 = P + x0    
    
    factor = 1.0/(x1*x3)
    
    x5 = _linstage0(x11,x21,x4)
    x6 = _linstage0(x11,x12,x1)
    x7 = _linstage0(x12,x22,x4)
    res = np.zeros_like(x11)
    for i in range(x11.shape[0]):
        res[i] = (x1*x11[i]*x3 - x3*x5[i] - (T + x2)*(x6[i] + x7[i] - x5[i]))*factor
    
    return res


class OnlineVariance(object):
    """USes the M2 algorithm to compute the variance in a streaming fashion"""

    def __init__(self):
        self.reset()

    def reset(self):
        self.count = 0.0
        self.wcount = 0.0
        self.wcount2=0.0
        self.mean = None
        self.M2 = None
       
    
    def update(self,value,weight=1.0):
        self.count+=1
        self.wcount+=weight
        self.wcount2+=weight*weight


        if self.mean is None:
            self.mean = value*0.0
            self.M2 = value*0.0
        

        mean_old = self.mean
        try:
            self.mean = mean_old + (weight / self.wcount) * (value - mean_old)
        except ZeroDivisionError:
            self.mean = value*0.0
        self.M2 += weight * (value - mean_old) * (value - self.mean)

    @property
    def variance(self):
        if self.count < 2:
            return np.nan
        else:
            return self.M2/self.wcount
    
    @property
    def sampleVariance(self):
        if self.count < 2:
            return np.nan
        else:
            return self.M2/(self.wcount-1)


    # def combine_variance(self,averages, variances, counts):
    #     good_idx = [idx for idx,a in enumerate(averages) if not test_nan(a)]
    #     averages = [averages[idx] for idx in good_idx]
    #     variances = [variances[idx] for idx in good_idx]
    #     counts = [counts[idx] for idx in good_idx]
    #     good_variance = None
    #     if not test_nan(variances):
    #         try:
    #             good_variance = variances[np.where(~np.isnan(variances))[0][0]]*0.0
    #         except IndexError:
    #             good_variance = None
    #     #print(good_idx,'Good',good_variance)
    #     variances = [v if not test_nan(v) else good_variance for v in variances] 
    #     #print('NEWAVERAGES',averages) 
    #     #print('NEW WEIGHTS',counts)      
        
    #     average = np.average(averages, weights=counts,axis=0)
        
    #     #print('final average',average)
    #     size = np.sum(counts)
        
    #     counts = np.array(counts) * size/np.sum(counts)
    #     if hasattr(average,'__len__'):
    #         average = average[None,...]
    #         for x in range(1,len(average.shape)):
    #             counts = counts[:,None]
    #     squares = 0.0
    #     if good_variance is not None:
    #         squares = counts*np.nan_to_num(variances)
    #     #print(counts,variances,squares)
    #     squares = squares + counts*(average - averages)**2

    #     return average,np.sum(squares,axis=0)/size

    def combine_variance(self, averages, variance, counts):
        average = None
        size = np.sum(counts)
        for avg,cnt in zip(averages,counts):
            if cnt == 0:
                continue

            #print('avg',avg)
            if avg is not None and not avg is np.nan:
                if average is None:
                    average = avg*cnt
                else:
                    average += avg*cnt
        average/=size
        #print('AVERGAE',average)
        counts = np.array(counts) * size/np.sum(counts)

        squares = None

        for avg,cnt,var in zip(averages,counts,variance):
            #print('COUNT ',cnt)
            if cnt == 0.0:
                continue
            if cnt > 0.0:
                if squares is None:
                    squares = cnt*(average - avg)**2
                else:
                    squares += cnt*(average - avg)**2
            if var is not np.nan:
                squares += cnt*var 
        # squares = counts*variances
        # squares += counts*(average - averages)**2

        return average,squares/size
    def parallelVariance(self):
        from taurex import mpi

        variance = self.variance
        
        mean = self.mean
        if mean is None:
            mean = np.nan


        variances = mpi.allgather(variance)


        averages = mpi.allgather(mean)
        counts = mpi.allgather(self.wcount)
        #all_data = [(v,m,c) for v,m,c in zip(variances,averages,counts) if not v is 0 and not averages is 0.0 and not counts is 0.0]        
        #print('VARIANCES',variances)
        #print('AVERAGES',averages)
        #print('COUNTS',counts)
        
        finalvariance = self.combine_variance(averages,variances,counts)
        return finalvariance[-1]

        
        




import os

import numpy as np
import h5py
from taurex import OutputSize
def store_taurex_results(output,model,native_grid,absp,tau,contributions,observed=None,optimizer=None):


    o = output.create_group('Output')


    if observed:
        obs = o.create_group('Observed')
        observed.write(obs)


    

    if optimizer:

        p = output.create_group('Parameters')
        store_planet(p, model)
        store_star(p, model)
        store_temperature(p, model)
        store_pressure(p, model)
        store_chemistry(p, model)

        opt = output.create_group('Optimizer')
        store_optimizer(opt, model, optimizer)

        #pr = o.create_group('Profiles')
        #store_profiles(pr, model)

        store_fit(o, model, optimizer)


    else:
        fm = o.create_group('Forward')
        sp = fm.create_group('Spectrum')
        co = fm.create_group('Contributions')
        pr = fm.create_group('Profiles')

        p = output.create_group('Parameters')

        store_profiles(pr, model)
        store_planet(p, model)
        store_star(p, model)
        store_temperature(p, model)
        store_pressure(p, model)
        store_chemistry(p, model)

        store_fwspectrum(sp, native_grid, absp, tau)
        store_fwcontrib(co, contributions)

    optimizer.add_data_from_solutions(output)


def store_profiles(output,model):
    output.write_array('density_profile',model.densityProfile)
    output.write_array('scaleheight_profile',model.scaleheight_profile)
    output.write_array('altitude_profile',model.altitudeProfile)
    output.write_array('gravity_profile',model.gravity_profile)
    output.write_array('pressure_profile', model.pressure.profile)
    output.write_array('temp_profile', model.temperatureProfile)
    #output.write_array('temp_profile', model._temperature_profile.profile)

    output.write_array('active_mix_profile', model.chemistry.activeGasMixProfile)
    output.write_array('inactive_mix_profile', model.chemistry.inactiveGasMixProfile)

def store_fwspectrum(output,native_grid,absp,tau):

    output.write_array('native_wngrid', native_grid)
    output.write_array('native_wlgrid', 10000/native_grid)
    output.write_array('native_spectrum', absp)
    output.write_array('native_tau', tau)

def store_fwcontrib(output,contributions):
    for name, value in contributions:
        output.write_array(name, value)


def store_optimizer(output,model,opt):
    opt.write_optimizer(output)


def store_fit(output, model, opt):
    opt.write_fit(output)


def store_planet(output,model):
    model._planet.write(output)
def store_star(output,model):
    model._star.write(output)

def store_temperature(output,model):
    model._temperature_profile.write(output)

def store_pressure(output,model):
    model._pressure_profile.write(output)

def store_chemistry(output,model):
    model._chemistry.write(output)



def generate_profile_dict(model):
    out = {}
    out['temp_profile']=model.temperatureProfile[...]
    out['active_mix_profile']=model.chemistry.activeGasMixProfile[...]
    out['inactive_mix_profile']=model.chemistry.inactiveGasMixProfile[...]
    out['density_profile']=model.densityProfile[...]
    out['scaleheight_profile']=model.scaleheight_profile[...]
    out['altitude_profile']=model.altitudeProfile[...]
    out['gravity_profile']=model.gravity_profile[...]
    out['pressure_profile']=model.pressure.profile[...]
    return out

def generate_spectra_dict(result, contrib_result, native_grid, bin_grid=None):
    from taurex.util import bindown
    out = {}

    

    #Store model output
    #out['binned_model'] = result[0] 
    out['native_spectrum'] = result[1]
    out['native_tau'] = result[2]
    out['native_wngrid']= native_grid
    out['native_wlgrid']= 10000/native_grid

    if bin_grid is not None:
        out['bin_wngrid']= bin_grid
        out['bin_wlgrid']= 10000 / bin_grid
        out['bin_spectrum']= result[0]

        if native_grid.shape[0] != result[2].shape[1]:
            native_grid = native_grid[(native_grid >= bin_grid.min()) & (native_grid <= bin_grid.max())]
        
        out['bin_tau']= bindown(native_grid,result[2],bin_grid)




    contributions = {}
    
    main_contrib = result[-1]

    if contrib_result is not None:
        
        for contrib_name,contrib_list in contrib_result.items(): #Loop through each contribtuion
            
            contributions[contrib_name] = {}

            for k,v in main_contrib[contrib_name].items():
                contributions[contrib_name][k] = v
            
            
            for c in contrib_list: #Loop through its components
                name = c[0]
                native = None
                binned = None
                tau = None
                extra = None
                #Cause I had no forsight
                if len(c) == 3:

                    native = c[1]
                    tau = c[2]
                elif len(c)>3:
                    if isinstance(c[3],tuple):
                        native = c[1]
                        tau=c[2]
                        extra = c[3:]
                    else:
                        binned = c[1]
                        native=c[2]
                        tau=c[3]
                        extra=c[4:]
                
                #tau = c[3] # necessary?
                contrib_comp = {}
                if binned is not None:
                    contrib_comp['binned'] = binned
                contrib_comp['native'] = native
                #contrib_comp['tau'] = tau
                if extra is not None:
                    for k,v in extra:
                        contrib_comp[k] = v

                contributions[contrib_name][name] = contrib_comp

    out['Contributions'] = contributions
    return out


def plot_taurex_results_from_hdf5(arg_output):

    file = h5py.File(arg_output,'r')



    solution_name = file['Output']['solutions']
    spectrum = np.zeros((len(solution_name['solution0']['Spectra']['bin_wngrid']),len(solution_name)*2))

    for s in range(len(solution_name)):
        spectrum[:,2 * s] = 10000/solution_name['solution{}'.format(s)]['Spectra']['bin_wngrid'][:]
        spectrum[:,2 * s + 1] = solution_name['solution{}'.format(s)]['Spectra']['bin_spectrum'][:]

    observed = np.zeros((len(file['Output']['Observed']['spectrum']), 4))
    observed[:,0] = file['Output']['Observed']['wlgrid'][:]
    observed[:, 1]= file['Output']['Observed']['spectrum'][:]
    observed[:, 2]= file['Output']['Observed']['errorbars'][:]
    observed[:, 3]= file['Output']['Observed']['binwidths'][:]

    plot_spectrum(spectrum, os.path.splitext(arg_output)[0] +'_spectrum.pdf', observed=observed)

def plot_spectrum(spectrum, arg_output, observed = None):
    import matplotlib.pyplot as plt
    from matplotlib import cm
    cmap = cm.get_cmap('Set1')


    plt.figure()
    for i in range(int(len(spectrum[0,:])/2)):
        plt.plot(spectrum[:, 2*i], spectrum[:, 2*i+1], label="Fit", color=cmap(float(i / 12.)))

    plt.plot(observed[:, 0], observed[:, 1], '.', color='blue', alpha=0.6, label="Observed")
    plt.plot([observed[:, 0], observed[:, 0]],
             [observed[:, 1] - observed[:, 2], observed[:, 1] + observed[:, 2]], '-',
             color='blue', alpha=0.6)
    plt.plot([observed[:, 0] - observed[:, 3] / 2, observed[:, 0] + observed[:, 3] / 2],
             [observed[:, 1], observed[:, 1]], '-', color='blue', alpha=0.6)

    plt.gca().set_xscale('log')
    plt.xlabel('Wavelength ($\mu$m)')
    plt.ylabel('$(R_p / R_s)^2$')
    plt.xticks([0.5, 1, 2, 5, 10], [0.5, 1, 2, 5, 10])
    plt.legend(loc='upper left')
    plt.savefig(os.path.join(arg_output), dpi=1000)


def store_contributions(binner,model,output_size=OutputSize.heavy):

    native_grid,contribs = model.model_contrib()
    native_grid,contribs_component = model.model_full_contrib()

    contribution_dict = {}

    for contrib_name,main_contribution in contribs.items():
        
        flux,tau,extras = main_contribution
        
        this_contrib_dict = binner.generate_spectrum_output((native_grid,flux,tau,extras),output_size=output_size)

        try:
            del this_contrib_dict['native_wngrid']
            del this_contrib_dict['native_wnwidth']
        except KeyError:
            pass
        
        try:
            del this_contrib_dict['native_wlgrid']
            del this_contrib_dict['native_wlwidth']
        except KeyError:
            pass

        try:
            del this_contrib_dict['binned_wngrid']
            del this_contrib_dict['binned_wnwidth']
        except KeyError:
            pass

        try:
            del this_contrib_dict['binned_wlgrid']
            del this_contrib_dict['binned_wlwidth']
        except KeyError:
            pass

        for name,flux,tau,extras in contribs_component[contrib_name]:
            

            component_contrib_dict = binner.generate_spectrum_output((native_grid,flux,tau,extras),output_size=output_size)

            try:
                del component_contrib_dict['native_wngrid']
                del component_contrib_dict['native_wnwidth']
            except KeyError:
                pass
            
            try:
                del component_contrib_dict['native_wlgrid']
                del component_contrib_dict['native_wlwidth']
            except KeyError:
                pass

            try:
                del component_contrib_dict['binned_wngrid']
                del component_contrib_dict['binned_wnwidth']
            except KeyError:
                pass

            try:
                del component_contrib_dict['binned_wlgrid']
                del component_contrib_dict['binned_wlwidth']
            except KeyError:
                pass

            this_contrib_dict[name] = component_contrib_dict

        contribution_dict[contrib_name] = this_contrib_dict
    
    return contribution_dict


#!/usr/bin/env python
import setuptools
from setuptools import find_packages
from numpy.distutils.core import setup
from numpy.distutils.core import Extension
from numpy.distutils import log
import re, os

packages = find_packages(exclude=('tests', 'doc'))
provides = ['taurex', ]


requires = []


install_requires = ['numpy',
                    'cython',
                    'configobj',
                    'scipy',
                    'numba',
                    'astropy',
                    'numexpr',
                    'numpy',
                    'nestle',
                    'h5py',
                    'tabulate', ]

console_scripts = ['taurex=taurex.taurex:main',
                   'taurex-plot=taurex.plot.plotter:main [Plot]']



def build_ace(parent_package='',top_path=None):
    from numpy.distutils.misc_util import Configuration

    ace_sources = ['taurex/external/ace.pyf',
    'src/ACE/Md_Types_Numeriques.f90',
                'src/ACE/Md_Constantes.f90',
                'src/ACE/Md_numerical_recipes.f90',
                'src/ACE/Md_Utilitaires.f90','src/ACE/Md_ACE.f90']

    data_files = ('taurex/external/ACE', ['src/ACE/Data/NASA.therm', 'src/ACE/Data/composes.dat'])
    ext = Extension(name='taurex.external.ace', sources=ace_sources)

    return ext, data_files

def build_bhmie():
    return Extension("taurex.external.mie",  
                        sources=["taurex/external/bh_mie.pyx",
                                "src/MIE/bhmie_lib.c",
                                "src/MIE/complex.c",
                                "src/MIE/nrutil.c"
                                ],
                        #extra_compile_args = [],
                        extra_compile_args=["-I./src/MIE/"],
                        language="c")

def _have_fortran_compiler():
    from numpy.distutils.fcompiler import available_fcompilers_for_platform, \
                                            new_fcompiler, \
                                            DistutilsModuleError, \
                                            CompilerNotFound
    from numpy.distutils import customized_fcompiler

    log.info('---------Detecting FORTRAN compilers-------')
    try:
        c = customized_fcompiler()
        v = c.get_version()
        return True
    except (DistutilsModuleError, CompilerNotFound, AttributeError) as e:
        return False

def _have_c_compiler():
    from distutils.errors import DistutilsExecError, DistutilsModuleError, \
                             DistutilsPlatformError, CompileError
    from numpy.distutils import customized_ccompiler
    log.info('---------Detecting C compilers-------')
    try:
        c = customized_ccompiler()
        v = c.get_version()
        return True
    except (DistutilsModuleError, CompileError, AttributeError) as e:
        return False

def create_extensions():
    try:
        from Cython.Build import cythonize
    except ImportError:
        log.warn('Could not import cython, ACE chemistry')
        log.warn('and BH Mie will not be installed')
        return [], []

    extensions = []
    data_files = []
    if _have_fortran_compiler():
        log.info('Detected FORTRAN compiler')
        log.info('ACE chemistry will be installed')

        ext, dat = build_ace()
        extensions.append(ext)
        data_files.append(dat)
    else:
        log.warn('No suitable FORTRAN compiler')
        log.warn('ACE chemistry will not be installed')

    if _have_c_compiler():
        log.info('Detected C compiler')
        log.info('BH Mie will be installed')
        extensions.append(build_bhmie())
    else:
        log.warn('No suitable C compiler')
        log.warn('BH Mie will not be installed')   

    if len(extensions) > 0:
        extensions = cythonize(extensions, language_level=3)

    return extensions, data_files


extensions, data_files = create_extensions()

entry_points = {'console_scripts': console_scripts, }

classifiers = [
    'Development Status :: 4 - Beta',
    'Environment :: Console',
    'Environment :: Win32 (MS Windows)',
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: BSD License',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: POSIX',
    'Operating System :: POSIX :: Linux',
    'Operating System :: Unix',
    'Operating System :: OS Independent',
    'Programming Language :: Python',
    'Topic :: Scientific/Engineering',
    'Topic :: Software Development :: Libraries',
]

# Handle versioning
version = '3.0.3-alpha'

with open("README.md", "r") as fh:
    long_description = fh.read()

try:
    setup(name='taurex',
        author='Ahmed Faris Al-Refaie',
        author_email='ahmed.al-refaie.12@ucl.ac.uk',
        license="BSD",
        version=version,
        description='TauREx 3 retrieval framework',
        classifiers=classifiers,
        packages=packages,
        long_description=long_description,
        url='https://github.com/ucl-exoplanets/TauREx3_public/',
        long_description_content_type="text/markdown",
        keywords = ['exoplanet','retrieval','taurex','taurex3','atmosphere','atmospheric'],
        include_package_data=True,
        entry_points=entry_points,
        provides=provides,
        requires=requires,
        install_requires=install_requires,
        extras_require={
            'Plot':  ["matplotlib"], },
        data_files=data_files,
        ext_modules=extensions
        )
except ext_errors as ex:
    print(str(ex))
    print("The C and/or FORTRAN extension could not be compiled")
    setup(name='taurex',
        author='Ahmed Faris Al-Refaie',
        author_email='ahmed.al-refaie.12@ucl.ac.uk',
        license="BSD",
        version=version,
        description='TauREx 3 retrieval framework',
        classifiers=classifiers,
        packages=packages,
        long_description=long_description,
        url='https://github.com/ucl-exoplanets/TauREx3_public/',
        long_description_content_type="text/markdown",
        keywords = ['exoplanet','retrieval','taurex','taurex3','atmosphere','atmospheric'],
        include_package_data=True,
        entry_points=entry_points,
        provides=provides,
        requires=requires,
        install_requires=install_requires,
        extras_require={
            'Plot':  ["matplotlib"], },
        data_files=data_files,
        )


from .tprofile import TemperatureProfile
import numpy as np
from taurex.data.fittable import fitparam


class Rodgers2000(TemperatureProfile):
    """
    Layer-by-layer temperature - pressure profile retrieval using dampening
    factor Introduced in Rodgers (2000): Inverse Methods for Atmospheric
    Sounding (equation 3.26). Featured in NEMESIS code (Irwin et al., 2008,
    J. Quant. Spec., 109, 1136 (equation 19)
    Used in all Barstow et al. papers.

    Parameters
    ----------
    temperature_layers : :obj:`list`
        Temperature in Kelvin per layer of pressure

    correlation_length : float
        In scaleheights, Line et al. 2013 sets this to 7, Irwin et al sets
        this to 1.5 may be left as free and Pressure dependent parameter later.

    covariance_matrix : :obj:`array` , optional
        User can supply their own covaraince matrix

    """

    def __init__(self, temperature_layers=[], correlation_length=5.0,
                 covariance_matrix=None):
        super().__init__('Rodgers2000')

        self._tp_corr_length = correlation_length
        self._covariance = covariance_matrix
        self._T_layers = np.array(temperature_layers)
        self.generate_temperature_fitting_params()

    def gen_covariance(self):
        """
        Generate the covariance matrix if None is supplied
        """
        h = self._tp_corr_length
        pres_prof = (self.pressure_profile)

        return np.exp(-1.0 * np.abs(np.log(pres_prof[:, None] /
                                           pres_prof[None, :])) / h)

    def correlate_temp(self, cov_mat):
        cov_mat_sum = np.sum(cov_mat, axis=0)
        weights = cov_mat[:, :]/cov_mat_sum[:, None]
        return weights.dot(self._T_layers)

    @property
    def profile(self):

        cov_mat = self._covariance
        if cov_mat is None:
            cov_mat = self.gen_covariance()

        return self.correlate_temp(cov_mat)

    @fitparam(param_name='correlation_length',
              param_latex='$C_{L}$',
              default_fit=False,
              default_bounds=[1.0, 10.0])
    def correlationLength(self):
        """
        Correlation length in scale heights
        """
        return self._tp_corr_length

    @correlationLength.setter
    def correlationLength(self, value):
        self._tp_corr_length = value

    def generate_temperature_fitting_params(self):
        """
        Generates the temperature fitting parameters for each layer of the
        atmosphere For a 4 layer atmosphere the fitting parameters generated
        are ``T_0``, ``T_1``, ``T_2`` and ``T_3``
        """

        bounds = [1e5, 1e3]
        for idx, val in enumerate(self._T_layers):
            point_num = idx+1
            param_name = 'T_{}'.format(point_num)
            param_latex = '$T_{%i}$' % point_num

            def read_point(self, idx=idx):
                return self._T_layers[idx]

            def write_point(self, value, idx=idx):
                self._T_layers[idx] = value

            fget_point = read_point
            fset_point = write_point
            default_fit = False
            self.add_fittable_param(param_name, param_latex, fget_point,
                                    fset_point, 'linear', default_fit, bounds)

    def write(self, output):
        temperature = super().write(output)

        cov_mat = self._covariance
        if cov_mat is None:
            cov_mat = self.gen_covariance()

        temperature.write_array('covariance_matrix', cov_mat)
        temperature.write_array('temperature_layers', self._T_layers)
        temperature.write_scalar('correlation_length', self._tp_corr_length)
        return temperature



from .contribution import Contribution
import numpy as np


class RayleighContribution(Contribution):
    """
    Computes contribution from Rayleigh scattering
    """

    def __init__(self):
        super().__init__('Rayleigh')

    def prepare_each(self, model, wngrid):
        """
        Computes the weighted opacity due to rayleigh
        scattering for any possible molecules within
        atmosphere.

        Parameters
        ----------
        model: :class:`~taurex.model.model.ForwardModel`
            Forward model

        wngrid: :obj:`array`
            Wavenumber grid

        Yields
        ------
        component: :obj:`tuple` of type (str, :obj:`array`)
            Name of scattering molecule and the weighted rayeligh opacity.


        """
        from taurex.util.scattering import rayleigh_sigma_from_name

        self._ngrid = wngrid.shape[0]
        self._nmols = 1
        self._nlayers = model.nLayers
        molecules = model.chemistry.activeGases + model.chemistry.inactiveGases

        for gasname in molecules:

            if np.max(model.chemistry.get_gas_mix_profile(gasname)) == 0.0:
                continue
            sigma = rayleigh_sigma_from_name(gasname, wngrid)

            if sigma is not None:
                final_sigma = sigma[None, :] * \
                    model.chemistry.get_gas_mix_profile(gasname)[:, None]
                self.sigma_xsec = final_sigma
                yield gasname, final_sigma



from taurex.log import Logger
from taurex.data.fittable import Fittable
from taurex.output.writeable import Writeable


class TemperatureProfile(Fittable, Logger, Writeable):
    """

    *Abstract Class*

    Defines temperature profile for an atmosphere

    Must define:

    - :func:`profile`

    Parameters
    ----------
    name : str
        Name used in logging

    """

    def __init__(self, name):
        Logger.__init__(self, name)
        Fittable.__init__(self)

    def initialize_profile(self, planet=None, nlayers=100,
                           pressure_profile=None):
        """
        Initializes the profile

        Parameters
        ----------
        planet: :class:`~taurex.data.planet.Planet`

        nlayers: int
            Number of layers in atmosphere

        pressure_profile: :obj:`array`
            Pressure at each layer of the atmosphere

        """
        self.nlayers = nlayers
        self.nlevels = nlayers+1
        self.pressure_profile = pressure_profile
        self.planet = planet

    @property
    def profile(self):
        """
        Must return a temperature profile at each layer of the atmosphere

        Returns
        -------
        temperature: :obj:`array`
            Temperature in Kelvin
        """
        raise NotImplementedError

    def write(self, output):
        temperature = output.create_group('Temperature')
        temperature.write_string('temperature_type', self.__class__.__name__)

        return temperature



from taurex.log import Logger
import numpy as np


class Opacity(Logger):
    """
    This is the base class for computing opactities

    """

    def __init__(self, name):
        super().__init__(name)

    @property
    def resolution(self):
        raise NotImplementedError

    @property
    def moleculeName(self):
        raise NotImplementedError

    @property
    def wavenumberGrid(self):
        raise NotImplementedError

    @property
    def temperatureGrid(self):
        raise NotImplementedError

    @property
    def pressureGrid(self):
        raise NotImplementedError

    def compute_opacity(self, temperature, pressure, wngrid=None):
        raise NotImplementedError

    def opacity(self, temperature, pressure, wngrid=None):

        if wngrid is None:
            wngrid_filter = slice(None)
        else:
            wngrid_filter = np.where((self.wavenumberGrid >= wngrid.min()) & (
                self.wavenumberGrid <= wngrid.max()))[0]

        orig = self.compute_opacity(temperature, pressure, wngrid_filter)

        if wngrid is None or np.array_equal(self.wavenumberGrid.take(wngrid_filter), wngrid):
            return orig
        else:
            # min_max =  (self.wavenumberGrid <= wngrid.max() ) & (self.wavenumberGrid >= wngrid.min())

            # total_bins = self.wavenumberGrid[min_max].shape[0]
            # if total_bins > wngrid.shape[0]:
            #     return np.append(np.histogram(self.wavenumberGrid,wngrid, weights=orig)[0]/np.histogram(self.wavenumberGrid,wngrid)[0],0)

            # else:
            return np.interp(wngrid, self.wavenumberGrid[wngrid_filter], orig)



"""
Contains the abstract class used by all collisionally induced absorption
objects.
"""

from taurex.log import Logger
import numpy as np


class CIA(Logger):
    """

    *Abstract class*

    This is the base class for collisionally induced absorption opacities.
    To function in Taurex3, it requires concrete implementations of:
        - :func:`wavenumberGrid`
        - :func:`compute_cia`
        - :func:`temperatureGrid`



    Parameters
    ----------
    name : str
        Name to use for logging

    pair_name : str
        pair of molecules this class represents. e.g. 'H2-H2' or 'H2-He'

    """

    def __init__(self, name, pair_name):
        super().__init__(name)

        self._pair_name = pair_name

    @property
    def pairName(self):
        """
        The assigned pair of molecules of this CIA

        Returns
        -------
        str
            The pair of molecules of this object in the form: ``Molecule1``-``Molecule2``

        """
        return self._pair_name


    @property
    def pairOne(self):
        """
        The name of the first molecule in the pair

        Returns
        -------
        str
            First molecule in the pair

        """

        return self._pair_name.split('-')[0]

    @property
    def pairTwo(self):
        """
        The name of the second molecule in the pair

        Returns
        -------
        str
            Second molecule in the pair

        """

        return self._pair_name.split('-')[-1]

    def compute_cia(self, temperature):
        """
        Computes the collisionaly induced cross-section for a given temeprature

        Unimplemented, this must be implemented in any derived class to be
        considered compatible in Taurex3

        The rules are:
            1. It must accept temperature in Kelvin (K)
            2. If the temperature falls outside of :func:`temperatureGrid` it must be set to zero
            3. The returned array must be of equal size to :func:`wavenumberGrid`


        Parameters
        ----------
        temperature : float
            Temeprature in Kelvin

        Returns
        -------
        :obj:`array`
            CIA cross section at desired temeprature on its native grid

        Raises
        ------
        NotImplementedError
            Only if derived class does not implement this

        """
        raise NotImplementedError

    @property
    def wavenumberGrid(self):
        """
        The native wavenumber grid (cm-1) of the CIA. Must be implemented in
        derived classes

        Returns
        -------
        :obj:`array`
            Native wavenumber grid

        Raises
        ------
        NotImplementedError
            Only if derived class does not implement this

        """
        raise NotImplementedError

    @property
    def temperatureGrid(self):

        """
        The native temperature grid of the CIA cross-sections.

        Returns
        -------
        :obj:`array`
            Native temeprature grid in Kelvin

        Raises
        ------
        NotImplementedError
            Only if derived class does not implement this

        """

        raise NotImplementedError

    def cia(self, temperature, wngrid=None):
        """
        For a given temperature, computes the appropriate cross section.
        If wavenumber grid ( :obj:`wngrid` ) is provided then the
        cross-section is interpolated
        to it.

        Parameters
        ----------
        temperature : float
            Temeprature in Kelvin

        wngrid : :obj:`array` , optional
            Wavenumber grid to interpolate to


        Returns
        -------
        :obj:`array`
            CIA cross section at desired temeprature on either its native grid
            or interpolated on :obj:`wngrid` if supplied

        """

        orig = self.compute_cia(temperature)
        if wngrid is None:
            return orig
        else:
            return np.interp(wngrid, self.wavenumberGrid, orig)




import numpy as np


def voigt_wolfz(nu, alpha, gamma):
    import numpy as np
    from scipy.special import wofz
    """
    Return the Voigt line shape at x with Lorentzian component HWHM gamma
    and Gaussian component HWHM alpha.

    """
    RT2LN2 = 1.1774100225154747
    RTPI   = 1.7724538509055159
    RT2PI  = 2.5066282746310002
    RT2    = 1.4142135623730951

    sigma = alpha / RT2LN2
    x = nu/ sigma / RT2
    y = gamma / sigma / RT2
    z = (x + 1j*gamma)

    return np.real(wofz(z)) /sigma/ RT2PI


    def compute_voigt(xgrid,transfreq,intens,dopplercoef,lorentz,cutoff=25.0):
        xsec = np.zeros_like(xgrid)
        start_wn = xgrid.searchsorted(transfreq-cutoff)
        end_wn = xgrid.searchsorted(transfreq+cutoff)-1
        mask_trans,mask_grid_flat = np.where( (xgrid[start_wn [:,None]] <=xgrid) & (xgrid[end_wn[:,None]] >= xgrid))
        doppler_width = lorentz#(transfreq*dopplercoef)
    
        vn_grid = transfreq[mask_trans]
        intens_grid = intens[mask_trans]
    
        doppl_grid = doppler_width[mask_trans]
        loren_grid = lorentz[mask_trans]
    
        abs_grid=np.abs(xgrid[mask_grid_flat]-vn_grid)
    
        voigt = voigt_wolfz(abs_grid, doppl_grid, loren_grid)*intens_grid
        s = np.argsort(mask_grid_flat)
        s_mask_grid = mask_grid_flat[s]
        s_voigt = voigt[s]
    
        uniq,index = np.unique(s_mask_grid,return_index=True)
    
    
        for idx,st,ed in zip(uniq[:-1],index[:-1],index[1:]):
            xsec[idx] += np.sum(s_voigt[st:ed])
    
        xsec[uniq[-1]] += np.sum(s_voigt[index[-1]:])
        return xsec


from .tprofile import TemperatureProfile
import numpy as np


class TemperatureArray(TemperatureProfile):
    """

    Temperature profile loaded from array

    """

    def __init__(self, tp_array=[2000, 1000]):
        super().__init__(self.__class__.__name__)

        self._tp_profile = np.array(tp_array)

    @property
    def profile(self):
        """Returns an isothermal temperature profile

        Returns: :obj:np.array(float)
            temperature profile
        """

        if self._tp_profile.shape[0] == self.nlayers:
            return self._tp_profile
        else:
            interp_temp = np.linspace(0.0, 1.0, self._tp_profile.shape[0])
            interp_array = np.linspace(0.0, 1.0, self.nlayers)

            return np.interp(interp_array, interp_temp, self._tp_profile)

    def write(self, output):

        temperature = super().write(output)

        temperature.write_scalar('tp_array', self._tp_profile)

        return temperature



from taurex.log import Logger
from taurex.data.fittable import Fittable
from taurex.output.writeable import Writeable


class Gas(Fittable, Logger, Writeable):
    """

    *Abstract Class*

    This class is a base for a single molecule or gas.
    Its used to describe how it mixes at each layer and combined
    with
    :class:`~taurex.data.profile.chemistry.taurexchemistry.TaurexChemistry`
    is used to build a chemical profile of the planets atmosphere.
    Requires implementation of:

    - func:`~mixProfile`


    Parameters
    -----------

    name :str
        Name used in logging

    molecule_name : str
        Name of molecule

    """

    def __init__(self, name, molecule_name):
        Logger.__init__(self, name)
        Fittable.__init__(self)
        self._molecule_name = molecule_name
        self.mix_profile = None

    @property
    def molecule(self):
        """
        Returns
        -------
        molecule_name: str
            Name of molecule
        """
        return self._molecule_name

    @property
    def mixProfile(self):
        """
        **Requires implementation**

        Should return mix profile of molecule/gas at each layer

        Returns
        -------
        mix: :obj:`array`
            Mix ratio for molecule at each layer

        """
        raise NotImplementedError

    def initialize_profile(self, nlayers=None, temperature_profile=None,
                           pressure_profile=None, altitude_profile=None):
        """
        Initializes and computes mix profile

        Parameters
        ----------
        nlayers: int
            Number of layers in atmosphere

        temperature_profile: :obj:`array`
            Temperature profile of atmosphere in K. Length must be
            equal to ``nlayers``

        pressure_profile: :obj:`array`
            Pressure profile of atmosphere in Pa. Length must be
            equal to ``nlayers``

        altitude_profile: :obj:`array`
            Altitude profile of atmosphere in m. Length must be
            equal to ``nlayers``

        """
        pass

    def write(self, output):
        """
        Writes class and arguments to file

        Parameters
        ----------
        output: :class:`~taurex.output.output.Output`

        """
        gas_entry = output.create_group(self.molecule)
        gas_entry.write_string('gas_type', self.__class__.__name__)
        gas_entry.write_string('molecule_name', self._molecule_name)
        return gas_entry



import configobj
from taurex.log import Logger
from .factory import *

class ParameterParser(Logger):

    def __init__(self):
        super().__init__('ParamParser')
        self._read = False
    

    def transform(self, section, key):
        val = section[key]
        newval = val
        if isinstance(val, list):
            try:
                newval = list(map(float,val))

            except:
                pass
        elif isinstance(val, (str)):
            if val.lower() in ['true',  'yes', 'yeah', 'yup', 'certainly', 'uh-huh',]:
                newval = True
            elif val.lower() in ['false',  'no', 'nope', 'no-way', 'hell-no',]:
                newval = False
            else:
                try:
                    newval = float(val)
                except:
                    pass
        section[key]=newval
        return newval


    def setup_globals(self):
        from taurex.cache import CIACache,OpacityCache
        config = self._raw_config.dict()
        if 'Global' in config:
            try:
                OpacityCache().set_opacity_path(config['Global']['xsec_path'])
            except KeyError:
                self.warning('No xsec path set, opacities cannot be used in model')
            try:
                
                OpacityCache().set_interpolation(config['Global']['xsec_interpolation'])
                self.info('Interpolation mode set to {}'.format(config['Global']['xsec_interpolation']))
            except KeyError:
                self.info('Interpolation mode set to linear')

            try:
                CIACache().set_cia_path(config['Global']['cia_path'])
            except KeyError:
                self.warning('No cia path set, cia cannot be used in model')
            
            try:
                OpacityCache().set_memory_mode(config['Global']['xsec_in_memory'])
            except KeyError:
                self.warning('Xsecs will be loaded in memory')
            
            try:
                OpacityCache().enable_radis(config['Global']['use_radis'])
            except KeyError:
                self.warning('Radis is disabled')

            try:
                wn_start,wn_end,wn_points = config['Global']['radis_grid']

                OpacityCache().set_radis_wavenumber(wn_start,wn_end,wn_points)
            except KeyError:
                self.warning('Radis default grid will be used')


    def read(self,filename):
        import os.path
        if not os.path.isfile(filename):
            raise Exception('Input file {} does not exist'.format(filename))
        self._raw_config = configobj.ConfigObj(filename)
        self.debug('Raw Config file is {}, filename is {}'.format(self._raw_config,filename))
        self._raw_config.walk(self.transform)
        config = self._raw_config.dict()
        self.debug('Config file is {}, filename is {}'.format(config,filename))

    def generate_lightcurve(self):
        config = self._raw_config.dict()
        if 'Lightcurve' in config:
            from taurex.model.lightcurve.lightcurve import LightCurveModel
            model = self.generate_model()
            lightcurvefile = config['Lightcurve']['lc_pickle']
            return LightCurveModel(model,lightcurvefile)
        else:
            raise KeyError

    
    def generate_appropriate_model(self):

        try:
            return self.generate_lightcurve()
        except KeyError:
            return self.generate_model()

    def generate_instrument(self, binner=None):

        config = self._raw_config.dict()
        if 'Instrument' in config:
            inst_config = config['Instrument']

            try:
                num_obs = inst_config.pop('num_observations')
            except KeyError:
                num_obs = 1

            if 'instrument' in inst_config:
                if inst_config['instrument'].lower() in ('snr', 'signalnoise', ):
                    inst = self.create_snr(binner, inst_config)
                    return inst,num_obs
            inst = create_instrument(inst_config)
            return inst,num_obs
        else:
            return None

    def create_snr(self,binner,config):
        from taurex.instruments.snr import SNRInstrument
        if binner is None:
            self.critical('Binning must be defined for SNR instrument')
            raise ValueError('Binning must be defined for SNR instrument')
        else:
            SNR = 10
            if 'SNR' in config:
                SNR = config['SNR']
            
            return SNRInstrument(SNR=SNR,binner=binner)

    def generate_optimizer(self):
        config = self._raw_config.dict()
        if 'Optimizer' in config:
            return create_optimizer(config['Optimizer'])
        else:
            None

    def generate_observation(self):

        config = self._raw_config.dict()
        if 'Observation' in config:
            observation_config = config['Observation']
            if 'lightcurve' in observation_config:
                from taurex.data.spectrum.lightcurve import ObservedLightCurve
                return ObservedLightCurve(observation_config['lightcurve'])

            elif 'observed_spectrum' in observation_config:
                from taurex.data.spectrum.observed import ObservedSpectrum
                return ObservedSpectrum(observation_config['observed_spectrum'])
            elif 'phasecurve_path' in observation_config:
                self.info('Phase curves to be implemented soon...... :D ')
            elif 'taurex_spectrum' in observation_config:
                if observation_config['taurex_spectrum'] == 'self':
                    return 'self'
                from taurex.data.spectrum.taurex import TaurexSpectrum
                return TaurexSpectrum(observation_config['taurex_spectrum'])
            elif 'iraclis_spectrum' in observation_config:
                from taurex.data.spectrum.iraclis import IraclisSpectrum
                return IraclisSpectrum(observation_config['iraclis_spectrum'])
            else:
                self.warning('No observation specified........')
                return None
        return None
    
    def create_manual_binning(self, config):
        import numpy as np
        import math
        from taurex.binning import FluxBinner, SimpleBinner
        
        binning_class = SimpleBinner

        if 'accurate' in config:
            if config['accurate']:
                binning_class = FluxBinner
        
        # Handle wavelength grid
        wngrid = None
        if 'wavelength_grid' in config:
            start, end, size = config['wavelength_grid']
            wngrid = 10000/np.linspace(start, end, int(size))
            wngrid = np.sort(wngrid)

        elif 'wavenumber_grid' in config:
            start, end, size = config['wavenumber_grid']
            wngrid = np.linspace(start, end, int(size))

        elif 'log_wavenumber_grid' in config:
            start, end, size = config['log_wavenumber_grid']
            start = math.log10(start)
            end = math.log10(end)
            wngrid = np.logspace(start, end, int(size))

        elif 'log_wavelength_grid' in config:
            start, end, size = config['log_wavelength_grid']
            start = math.log10(start)
            end = math.log10(end)
            wngrid = np.sort(10000/np.logspace(start, end, int(size)))
        elif 'wavelength_res' in config:
            from taurex.util.util import create_grid_res
            start, end, res = config['wavelength_res']
            wlgrid = create_grid_res(res, start, end)[:, 0].flatten()
            wngrid = 10000/wlgrid[::-1]

        if wngrid is None:
            self._logger.error('manual was selected and no grid was given.'
                               '(Use wavelength_grid, wavenumber_grid or log versions)')
            raise Exception('manual selected but no grid given')

        return binning_class(wngrid), wngrid

    def generate_binning(self):

        config = self._raw_config.dict()
        if 'Binning' in config:
            binning_config = config['Binning']
            if 'bin_type' in binning_config:
                bin_type = binning_config['bin_type'].lower()

                if bin_type == 'native':
                    return 'native'
                elif bin_type == 'observed':
                    return 'observed'
                elif bin_type == 'manual':
                    return self.create_manual_binning(binning_config)
            else:
                return None
        return None
    # def generate_spectrum(self):
    #     import numpy as np

    #     config = self._raw_config.dict()
    #     observed = None
    #     if 'Spectrum' in config:
    #         spectrum_config = config['Spectrum']
    #         if 'lightcurve' in spectrum_config:
    #             from taurex.data.spectrum.lightcurve import ObservedLightCurve
    #             observed = ObservedLightCurve(spectrum_config['lightcurve'])

    #         elif 'observed_spectrum' in spectrum_config:
    #             from taurex.data.spectrum.observed import ObservedSpectrum
    #             observed = ObservedSpectrum(spectrum_config['observed_spectrum'])

    #         if 'grid_type' in spectrum_config:
    #             grid_type = spectrum_config['grid_type']

    #             if grid_type == 'observed':
    #                 if observed is not None:
    #                     return observed,observed.wavenumberGrid
    #                 else:
    #                     self.critical('grid type is observed yet no observed_spectrum is defined!!!')
    #                     raise Exception('No observed spectrum defined for observed grid_type')
    #             elif grid_type == 'native':
    #                 return observed,None
    #             elif grid_type == 'manual':

    #                 if 'wavenumber_grid' in spectrum_config:
    #                     start,end,size = spectrum_config['wavenumber_grid']
    #                     return observed,np.linspace(start,end,int(size))
    #                 elif 'wavelength_grid' in spectrum_config:
    #                     start,end,size = spectrum_config['wavelength_grid']
    #                     return observed,np.linspace(10000/end,10000/start,int(size))
    #                 else:
    #                    self.critical('grid type is manual yet neither wavelength_grid or wavenumber_grid is defined')
    #                    raise Exception('wavenumber_grid/wavelength_grid not defined in input for manual grid_type')
    #             else:
    #                 return observed,None

    def generate_model(self, chemistry=None, pressure=None,
                       temperature=None, planet=None,
                       star=None):
        config = self._raw_config.dict()
        if 'Model' in config:
            if chemistry is None:
                chemistry = self.generate_chemistry_profile()
            if pressure is None:
                pressure = self.generate_pressure_profile()
            if temperature is None:
                temperature = self.generate_temperature_profile()
            if planet is None:
                planet = self.generate_planet()
            if star is None:
                star = self.generate_star()
            model= create_model(config['Model'],chemistry,temperature,pressure,planet,star)
        else:
            return None
        
        return model
    def generate_chemistry_profile(self):
        config = self._raw_config.dict()
        if 'Chemistry' in config:
            return create_chemistry(config['Chemistry'])
        else:
            return None

    def generate_pressure_profile(self):
        config = self._raw_config.dict()
        if 'Pressure' in config:
            return create_pressure_profile(config['Pressure'])
        else:
            return None
    
    def generate_temperature_profile(self):
        config = self._raw_config.dict()
        if 'Temperature' in config:
            return create_temperature_profile(config['Temperature'])
        else:
            return None
    
    def generate_planet(self):
        config = self._raw_config.dict()

        if 'Planet' in config:
            return create_planet(config['Planet'])
        else:
            return None
    def generate_star(self):
        config = self._raw_config.dict()

        if 'Star' in config:
            return create_star(config['Star'])
        else:
            return None


    def generate_fitting_parameters(self):
        config = self._raw_config.dict()
        if 'Fitting' in config:
            fitting_config = config['Fitting']

            fitting_params = {}

            for key,value in fitting_config.items():
                fit_param,fit_type=key.split(':')
                if not fit_param in fitting_params:
                    fitting_params[fit_param] = {'fit':None,'bounds':None,'mode':None,'factor':None}
                fitting_params[fit_param][fit_type]=value

        return fitting_params







from taurex.log import Logger
from taurex.data.fittable import fitparam, Fittable
import numpy as np
from taurex.output.writeable import Writeable


class PressureProfile(Fittable, Logger, Writeable):
    """
    *Abstract Class*

    Base pressure class. Simple. Defines the layering
    of the atmosphere. Only requires
    implementation of:

    - :func:`compute_pressure_profile`
    - :func:`profile`

    Parameters
    ----------

    name: str
        Name used in logging

    nlayers: int
        Number of layers in atmosphere

    """

    def __init__(self, name, nlayers):
        Fittable.__init__(self)
        Logger.__init__(self, name)

        self._nlayers = int(nlayers)

    @property
    def nLayers(self):
        """
        Number of layers

        Returns
        -------
        int
        """

        return self._nlayers

    @property
    def nLevels(self):
        return self.nLayers + 1

    def compute_pressure_profile(self):
        """
        **Requires implementation**

        Compute pressure profile and
        generate pressure array in Pa

        Returns
        -------
        pressure_profile: :obj:`array`
            Pressure profile array in Pa

        """
        raise NotImplementedError

    @property
    def profile(self):
        """
        Returns pressure at each atmospheric layer (Pascal)

        Returns
        -------
        pressure_profile : :obj:`array`
        """
        raise NotImplementedError

    def write(self, output):
        pressure = output.create_group('Pressure')
        pressure.write_string('pressure_type', self.__class__.__name__)
        pressure.write_scalar('nlayers', self._nlayers)
        pressure.write_array('profile', self.profile)
        return pressure


class SimplePressureProfile(PressureProfile):
    """
    A basic pressure profile.

    Parameters
    ----------
    nlayers : int
        Number of layers in atmosphere

    atm_min_pressure : float
        minimum pressure in Pascal (top of atmosphere)

    atm_max_pressure : float
        maximum pressure in Pascal (surface of planet)

    """

    def __init__(self, nlayers=100,
                 atm_min_pressure=1e-4,
                 atm_max_pressure=1e6):

        super().__init__('pressure_profile', nlayers)
        self.pressure_profile = None
        self._atm_min_pressure = atm_min_pressure
        self._atm_max_pressure = atm_max_pressure

    def compute_pressure_profile(self):
        """
        Sets up the pressure profile for the atmosphere model

        """

        # set pressure profile of layer boundaries
        press_exp = np.linspace(np.log(self._atm_min_pressure),
                                np.log(self._atm_max_pressure),
                                self.nLevels)
        self.pressure_profile_levels = np.exp(press_exp)[::-1]

        # get mid point pressure between levels (i.e. get layer pressure)
        # computing geometric
        # average between pressure at n and n+1 level
        self.pressure_profile = \
            np.power(10, np.log10(self.pressure_profile_levels)[:-1] +
                     np.diff(np.log10(self.pressure_profile_levels))/2.)

    @fitparam(param_name='atm_min_pressure',
              param_latex='$P_\mathrm{min}$',
              default_mode='log',
              default_fit=False,
              default_bounds=[0.1, 1.0])
    def minAtmospherePressure(self):
        """Minimum pressure of atmosphere (top layer) in Pascal"""
        return self._atm_min_pressure

    @minAtmospherePressure.setter
    def minAtmospherePressure(self, value):
        self._atm_min_pressure = value

    @fitparam(param_name='atm_max_pressure',
              param_latex='$P_\\mathrm{max}$',
              default_mode='log',
              default_fit=False,
              default_bounds=[0.1, 1.0])
    def maxAtmospherePressure(self):
        """Maximum pressure of atmosphere (surface) in Pascal"""

        return self._atm_max_pressure

    @maxAtmospherePressure.setter
    def maxAtmospherePressure(self, value):
        self._atm_max_pressure = value

    @property
    def profile(self):
        return self.pressure_profile

    def write(self, output):
        pressure = super().write(output)

        pressure.write_scalar('atm_max_pressure', self._atm_max_pressure)
        pressure.write_scalar('atm_min_pressure', self._atm_min_pressure)

        return pressure



import logging
__all__ = ['Logger']


root_logger = logging.getLogger('taurex')
root_logger.propagate = False
"""Root logger for taurex"""


class TauRexHandler(logging.StreamHandler):
    """
    Logging Handler for Taurex 3. Prevents other
    MPI threads from writing to log unless they are in trouble (>=ERROR)

    Parameters
    ----------
    stream : stream-object , optional
        Stream to write to otherwise defaults to ``stderr``

    """

    def __init__(self, stream=None):
        from taurex.mpi import get_rank
        super().__init__(stream=stream)

        self._rank = get_rank()

    def emit(self, record):
        # print(record)
        if self._rank == 0 or record.levelno >= logging.ERROR:
            # msg = '[{}] {}'.format(self._rank,record.msg)
            # record.msg = msg
            return super(TauRexHandler, self).emit(record)
        else:
            pass


rh = TauRexHandler()
formatter = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
rh.setFormatter(formatter)
rh.setLevel(logging.DEBUG)
root_logger.handlers = []
root_logger.addHandler(rh)
root_logger.setLevel(logging.INFO)


class Logger:
    """
    Standard logging using logger library

    Parameters
    -----------
    name : str
        Name used for logging

    """

    def __init__(self, name):
        self._log_name = 'taurex.{}'.format(name)

        self._logger = logging.getLogger('taurex.{}'.format(name))

    def info(self, message, *args, **kwargs):
        """ See :class:`logging.Logger` """
        self._logger.info(message, *args, **kwargs)

    def warning(self, message, *args, **kwargs):
        """ See :class:`logging.Logger` """
        self._logger.warning(message, *args, **kwargs)

    def debug(self, message, *args, **kwargs):
        """ See :class:`logging.Logger` """
        self._logger.debug(message, *args, **kwargs)

    def error(self, message, *args, **kwargs):
        """ See :class:`logging.Logger` """
        self._logger.error(message, *args, **kwargs)

    def critical(self, message, *args, **kwargs):
        """ See :class:`logging.Logger` """
        self._logger.critical(message, *args, **kwargs)



from .chemistry import Chemistry
import numpy as np
from taurex.util import molecule_texlabel
from taurex.cache import OpacityCache
from taurex.exceptions import InvalidModelException


class InvalidChemistryException(InvalidModelException):
    """
    Exception that is called when atmosphere mix is greater
    than unity
    """
    pass


class TaurexChemistry(Chemistry):

    """
    The standard chemical model used in Taurex. This allows for the combination
    of different mixing profiles for each molecule. Lets take an example
    profile, we want an atmosphere with a constant mixing of ``H2O`` but two
    layer mixing for ``CH4``.
    First we initialize our chemical model:

        >>> chemistry = TaurexChemistry()

    Then we can add our molecules using the :func:`addGas` method. Lets start
    with ``H2O``, since its a constant profile for all layers of the atmosphere
    we thus add
    the :class:`~taurex.data.profiles.chemistry.gas.constantgas.ConstantGas`
    object:

        >>> chemistry.addGas(ConstantGas('H2O',mix_ratio = 1e-4))

    Easy right? Now the same goes for ``CH4``, we can add the molecule into
    the chemical model by using the correct profile (in this case
    :class:`~taurex.data.profiles.chemistry.gas.twolayergas.TwoLayerGas`):

        >>> chemistry.addGas(TwoLayerGas('CH4',mix_ratio_surface=1e-4,
                                         mix_ratio_top=1e-8))

    Molecular profiles available are:
        * :class:`~taurex.data.profiles.chemistry.gas.constantgas.ConstantGas`
        * :class:`~taurex.data.profiles.chemistry.gas.twolayergas.TwoLayerGas`
        * :class:`~taurex.data.profiles.chemistry.gas.twolayergas.TwoPointGas`


    Parameters
    ----------

    fill_gases : str or :obj:`list`
        Either a single gas or list of gases to fill the atmosphere with

    ratio : float or :obj:`list`
        If a bunch of molecules are used to fill an atmosphere, whats the
        ratio between them?
        The first fill gas is considered the main one with others defined as
        ``molecule / main_molecule``


    """

    def __init__(self, fill_gases=['H2', 'He'], ratio=0.17567):
        super().__init__('ChemistryModel')

        self._gases = []
        self._active = []
        self._inactive = []

        if isinstance(fill_gases, str):
            fill_gases = [fill_gases]

        if isinstance(ratio, float):
            ratio = [ratio]

        if len(fill_gases) > 1 and len(ratio) != len(fill_gases)-1:
            self.error('Fill gases and ratio count are not correctly matched')
            self.error('There should be %s ratios, you have defined %s',
                       len(fill_gases)-1, len(ratio))
            raise InvalidChemistryException

        self._fill_gases = fill_gases
        self._fill_ratio = ratio
        self.active_mixratio_profile = None
        self.inactive_mixratio_profile = None
        self.molecules_i_have = OpacityCache().find_list_of_molecules()
        self.debug('MOLECULES I HAVE %s', self.molecules_i_have)
        self.setup_fill_params()

    def setup_fill_params(self):
        if not hasattr(self._fill_gases, '__len__') or \
                len(self._fill_gases) < 2:
            return

        main_gas = self._fill_gases[0]

        for idx, value in enumerate(zip(self._fill_gases[1:],
                                        self._fill_ratio)):
            gas, ratio = value
            mol_name = '{}_{}'.format(gas, main_gas)
            param_name = mol_name
            param_tex = '{}/{}'.format(molecule_texlabel(gas),
                                       molecule_texlabel(main_gas))

            def read_mol(self, idx=idx):
                return self._fill_ratio[idx]

            def write_mol(self, value, idx=idx):
                self._fill_ratio[idx] = value

            fget = read_mol
            fset = write_mol

            bounds = [1.0e-12, 0.1]

            default_fit = False
            self.add_fittable_param(param_name, param_tex, fget,
                                    fset, 'log', default_fit, bounds)

    def isActive(self, gas):
        """
        Determines if the gas is active or not (Whether we have cross-sections)

        Parameters
        ----------

        gas: str
            Name of molecule


        Returns
        -------
        bool:
            True if active
        """
        if gas in self.molecules_i_have:
            return True
        else:
            return False

    def addGas(self, gas):
        """
        Adds a gas in the atmosphere.

        Parameters
        ----------
        gas : :class:`~taurex.data.profiles.chemistry.gas.gas.Gas`
            Gas to add into the atmosphere. Only takes effect
            on next initialization call.

        """

        if gas.molecule in [x.molecule for x in self._gases]:
            self.error('Gas already exists %s', gas.molecule)
            raise InvalidChemistryException

        self.debug('Gas %s fill gas: %s', gas.molecule, self._fill_gases)
        if gas.molecule in self._fill_gases:
            self.error('Gas %s is already a fill gas: %s', gas.molecule,
                       self._fill_gases)
            raise InvalidChemistryException

        self._gases.append(gas)

    @property
    def activeGases(self):
        return self._active

    @property
    def inactiveGases(self):
        return self._inactive

    def fitting_parameters(self):
        """
        Overrides the fitting parameters to return
        one with all the gas profile parameters as well

        Returns
        -------

        fit_param : :obj:`dict`

        """
        full_dict = {}
        for gas in self._gases:
            full_dict.update(gas.fitting_parameters())

        full_dict.update(self._param_dict)

        return full_dict

    def initialize_chemistry(self, nlayers=100, temperature_profile=None,
                             pressure_profile=None, altitude_profile=None):
        """

        Initializes the chemical model and computes the all gas profiles
        and the mu profile for the forward model

        """
        self.info('Initializing chemistry model')

        active_profile = []
        inactive_profile = []

        self._active = []
        self._inactive = []
        for gas in self._gases:
            gas.initialize_profile(nlayers, temperature_profile,
                                   pressure_profile, altitude_profile)
            if self.isActive(gas.molecule):
                active_profile.append(gas.mixProfile)
                self._active.append(gas.molecule)
            else:
                inactive_profile.append(gas.mixProfile)
                self._inactive.append(gas.molecule)

        total_mix = sum(active_profile) + sum(inactive_profile)

        self.debug('Total mix output %s', total_mix)

        validity = np.any(total_mix > 1.0)

        self.debug('Is invalid? %s', validity)

        if validity:
            self.error('Greater than 1.0 chemistry profile detected')
            raise InvalidChemistryException

        mixratio_remainder = 1. - total_mix

        mixratio_remainder += np.zeros(shape=(nlayers))

        remain_active, remain_inactive = \
            self.fill_atmosphere(mixratio_remainder)

        active_profile.extend(remain_active)

        inactive_profile.extend(remain_inactive)

        if len(active_profile) > 0:
            self.active_mixratio_profile = np.vstack(active_profile)
        else:
            self.active_mixratio_profile = 0.0
        if len(inactive_profile) > 0:
            self.inactive_mixratio_profile = np.vstack(inactive_profile)
        else:
            self.inactive_mixratio_profile = 0.0
        super().initialize_chemistry(nlayers, temperature_profile,
                                     pressure_profile, altitude_profile)

    def fill_atmosphere(self, mixratio_remainder):
        active_profile = []
        inactive_profile = []

        if len(self._fill_gases) == 1:
            if self.isActive(self._fill_gases[0]):
                active_profile.append(mixratio_remainder)
                self._active.append(self._fill_gases[0])
            else:
                inactive_profile.append(mixratio_remainder)
                self._inactive.append(self._fill_gases[0])
        else:
            main_molecule = mixratio_remainder/(1. + sum(self._fill_ratio))
            if self.isActive(self._fill_gases[0]):
                active_profile.append(main_molecule)
                self._active.append(self._fill_gases[0])
            else:
                inactive_profile.append(main_molecule)
                self._inactive.append(self._fill_gases[0])
            for molecule, ratio in zip(self._fill_gases[1:], self._fill_ratio):
                second_molecule = ratio * main_molecule

                if self.isActive(molecule):
                    active_profile.append(second_molecule)
                    self._active.append(molecule)
                else:
                    inactive_profile.append(second_molecule)
                    self._inactive.append(molecule)
        return active_profile, inactive_profile

    @property
    def activeGasMixProfile(self):
        """
        Active gas layer by layer mix profile

        Returns
        -------
        active_mix_profile : :obj:`array`

        """
        return self.active_mixratio_profile

    @property
    def inactiveGasMixProfile(self):
        """
        Inactive gas layer by layer mix profile

        Returns
        -------
        inactive_mix_profile : :obj:`array`

        """
        return self.inactive_mixratio_profile

    def write(self, output):
        gas_entry = super().write(output)
        if isinstance(self._fill_gases, float):
            gas_entry.write_scalar('ratio', self._fill_ratio)
        elif hasattr(self._fill_gases, '__len__'):
            gas_entry.write_array('ratio', np.array(self._fill_ratio))
        gas_entry.write_string_array('fill_gases', self._fill_gases)
        for gas in self._gases:
            gas.write(gas_entry)

        return gas_entry



import h5py
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import numpy as np
import taurex.plot.corner as corner
import matplotlib as mpl
from taurex.util.util import decode_string_array
import os

from matplotlib import rc

# some global matplotlib vars
mpl.rcParams['axes.linewidth'] = 1  #set the value globally
mpl.rcParams['text.antialiased'] = True
mpl.rcParams['errorbar.capsize'] = 2

# rc('text', usetex=True) # use tex in plots
#rc('font', **{ 'family' : 'serif','serif':['Palatino'], 'size'   : 11})

class Plotter(object):
    phi = 1.618

    modelAxis = {
        'TransmissionModel' : '$(R_p/R_*)^2$',
        'EmissionModel' : '$F_p/F_*$',
        'DirectImageModel' : '$F_p$'

    }

    def __init__(self,filename,title=None,prefix=None,cmap='Paired',out_folder='.'):
        self.fd = h5py.File(filename,'r')
        self.title = title
        self.cmap = mpl.cm.get_cmap(cmap)
        self.prefix=prefix
        if self.prefix is None:
            self.prefix = "output" 
        self.out_folder=out_folder

        if not os.path.exists(self.out_folder):
            os.makedirs(self.out_folder)

    @property
    def num_solutions(self,fd_position='Output'):
        return len([(int(k[8:]),v) for k,v in self.fd[fd_position]['Solutions'].items() if 'solution' in k])

    def solution_iter(self,fd_position='Output'):
        for idx,solution in [(int(k[8:]),v) for k,v in self.fd[fd_position]['Solutions'].items() if 'solution' in k]:
            yield idx,solution

    def forward_output(self):
        return self.fd['Output']

    def compute_ranges(self):

        solution_ranges = []



        mu_derived = None
        for idx,sol in self.solution_iter():
            
            mu_derived = self.get_mu_parameters(sol)


            fitting_names = self.fittingNames


            fit_params = sol['fit_params']
            param_list = []
            for fit_names in self.fittingNames:
                param_values = fit_params[fit_names]
                sigma_m = param_values['sigma_m'][()]
                sigma_p = param_values['sigma_p'][()]
                val = param_values['value'][()]

                param_list.append([val,val- 5.0*sigma_m,val+5.0*sigma_p])
            
            if mu_derived is not None:
                sigma_m = mu_derived['sigma_m'][()]
                sigma_p = mu_derived['sigma_p'][()]
                val = mu_derived['value'][()]     
                param_list.append([val, val- 5.0*sigma_m,val+5.0*sigma_p])
            
            solution_ranges.append(param_list)


        fitting_boundary_low = self.fittingBoundaryLow
        fitting_boundary_high = self.fittingBoundaryHigh

        if mu_derived is not None:
            fitting_boundary_low = np.concatenate((fitting_boundary_low, [-1e99]))
            fitting_boundary_high = np.concatenate((fitting_boundary_high, [1e99]))


        range_all = np.array(solution_ranges)

        range_min = np.min(range_all[:,:,1],axis=0)
        range_max = np.max(range_all[:,:,2],axis=0)

        range_min = np.where(range_min < fitting_boundary_low, fitting_boundary_low,range_min)
        range_max = np.where(range_max > fitting_boundary_high, fitting_boundary_high,range_max)

        return list(zip(range_min,range_max)) 
            

    @property
    def activeGases(self):
        return decode_string_array(self.fd['ModelParameters']['Chemistry']['active_gases'])

    @property
    def inactiveGases(self):
        return decode_string_array(self.fd['ModelParameters']['Chemistry']['inactive_gases'])


    def plot_fit_xprofile(self):

        for solution_idx, solution_val in self.solution_iter():

            fig = plt.figure(figsize=(7,7/self.phi))
            ax = fig.add_subplot(111)
            num_moles = len(self.activeGases+self.inactiveGases)

            profiles = solution_val['Profiles']
            pressure_profile = profiles['pressure_profile'][:]/1e5
            active_profile = profiles['active_mix_profile'][...]
            active_profile_std = profiles['active_mix_profile_std'][...]

            inactive_profile = profiles['inactive_mix_profile'][...]
            inactive_profile_std = profiles['inactive_mix_profile_std'][...]

            cols_mol = {}
            for mol_idx,mol_name in enumerate(self.activeGases):
                cols_mol[mol_name] = self.cmap(mol_idx/num_moles)

                prof = active_profile[mol_idx]
                prof_std = active_profile_std[mol_idx]

                plt.plot(prof,pressure_profile,color=cols_mol[mol_name], label=mol_name)

                plt.fill_betweenx(pressure_profile, prof + prof_std, prof,
                                  color=self.cmap(mol_idx / num_moles), alpha=0.5)
                plt.fill_betweenx(pressure_profile, prof,
                                  np.power(10, (np.log10(prof) - (
                                              np.log10(prof + prof_std) - np.log10(prof)))),
                                  color=self.cmap(mol_idx / num_moles), alpha=0.5)

            for mol_idx,mol_name in enumerate(self.inactiveGases):
                inactive_idx = len(self.activeGases) + mol_idx
                cols_mol[mol_name] = self.cmap(inactive_idx/num_moles)

                
                prof = inactive_profile[mol_idx]
                prof_std = inactive_profile_std[mol_idx]

                plt.plot(prof,pressure_profile,color=cols_mol[mol_name], label=mol_name)

                plt.fill_betweenx(pressure_profile, prof + prof_std, prof,
                                  color=self.cmap(inactive_idx / num_moles), alpha=0.5)
                plt.fill_betweenx(pressure_profile, prof,
                                  np.power(10, (np.log10(prof) - (
                                              np.log10(prof + prof_std) - np.log10(prof)))),
                                  color=self.cmap(inactive_idx / num_moles), alpha=0.5)

        plt.yscale('log')
        plt.gca().invert_yaxis()
        plt.xscale('log')
        plt.xlim(1e-12, 3)
        plt.xlabel('Mixing ratio')
        plt.ylabel('Pressure (bar)')
        plt.tight_layout()
        box = ax.get_position()
        ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])
        ax.legend(loc='center left', bbox_to_anchor=(1, 0.5), ncol=1, prop={'size':11}, frameon=False)
        if self.title:
            plt.title(self.title, fontsize=14)
        plt.savefig(os.path.join(self.out_folder, '%s_fit_mixratio.pdf' % (self.prefix)))
        plt.close('all')

    def plot_forward_xprofile(self):
        fig = plt.figure(figsize=(7,7/self.phi))
        ax = fig.add_subplot(111)
        num_moles = len(self.activeGases+self.inactiveGases)

        solution_val = self.forward_output()

        profiles = solution_val['Profiles']
        pressure_profile = profiles['pressure_profile'][:]/1e5
        active_profile = profiles['active_mix_profile'][...]

        inactive_profile = profiles['inactive_mix_profile'][...]

        cols_mol = {}
        for mol_idx,mol_name in enumerate(self.activeGases):
            cols_mol[mol_name] = self.cmap(mol_idx/num_moles)

            prof = active_profile[mol_idx]

            plt.plot(prof,pressure_profile,color=cols_mol[mol_name], label=mol_name)

        for mol_idx,mol_name in enumerate(self.inactiveGases):
            inactive_idx = len(self.activeGases) + mol_idx
            cols_mol[mol_name] = self.cmap(inactive_idx/num_moles)

            
            prof = inactive_profile[mol_idx]

            plt.plot(prof,pressure_profile,color=cols_mol[mol_name], label=mol_name)

        plt.yscale('log')
        plt.gca().invert_yaxis()
        plt.xscale('log')
        plt.xlim(1e-12, 3)
        plt.xlabel('Mixing ratio')
        plt.ylabel('Pressure (bar)')
        plt.tight_layout()
        box = ax.get_position()
        ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])
        ax.legend(loc='center left', bbox_to_anchor=(1, 0.5), ncol=1, prop={'size':11}, frameon=False)
        if self.title:
            plt.title(self.title, fontsize=14)
        plt.savefig(os.path.join(self.out_folder, '%s_fit_mixratio.pdf' % (self.prefix)))
        plt.close('all')

    def plot_fitted_tp(self):

        # fitted model
        fig = plt.figure(figsize=(5,3.5))
        ax = fig.add_subplot(111)
        
        for solution_idx, solution_val in self.solution_iter():
            if self.num_solutions > 1:
                label = 'Fitted profile (%i)' % (solution_idx)
            else:
                label = 'Fitted profile'
            temp_prof = solution_val['Profiles']['temp_profile'][:]
            temp_prof_std = solution_val['Profiles']['temp_profile_std'][:]
            pres_prof = solution_val['Profiles']['pressure_profile'][:]/1e5
            plt.plot(temp_prof, pres_prof, color=self.cmap(float(solution_idx)/self.num_solutions), label=label)
            plt.fill_betweenx(pres_prof,  temp_prof-temp_prof_std,  temp_prof+temp_prof_std, color=self.cmap(float(solution_idx)/self.num_solutions), alpha=0.5)

        plt.yscale('log')
        plt.gca().invert_yaxis()
        plt.xlabel('Temperature (K)')
        plt.ylabel('Pressure (bar)')
        plt.tight_layout()
        legend = plt.legend(loc='upper left', ncol=1, prop={'size':11})
        legend.get_frame().set_facecolor('white')
        legend.get_frame().set_edgecolor('white')

        legend.get_frame().set_alpha(0.8)
        if self.title:
            plt.title(self.title, fontsize=14)
        plt.savefig(os.path.join(self.out_folder, '%s_tp_profile.pdf' % (self.prefix)))
        plt.close()

    def plot_forward_tp(self):

        fig = plt.figure(figsize=(5,3.5))
        ax = fig.add_subplot(111)
        
        solution_val = self.forward_output()

        temp_prof = solution_val['Profiles']['temp_profile'][:]
        pres_prof = solution_val['Profiles']['pressure_profile'][:]/1e5
        plt.plot(temp_prof, pres_prof)

        plt.yscale('log')
        plt.gca().invert_yaxis()
        plt.xlabel('Temperature (K)')
        plt.ylabel('Pressure (bar)')
        plt.tight_layout()
        legend = plt.legend(loc='upper left', ncol=1, prop={'size':11})
        legend.get_frame().set_facecolor('white')
        legend.get_frame().set_edgecolor('white')

        legend.get_frame().set_alpha(0.8)
        if self.title:
            plt.title(self.title, fontsize=14)
        plt.savefig(os.path.join(self.out_folder, '%s_tp_profile.pdf' % (self.prefix)))
        plt.close()


    def get_mu_parameters(self, solution):
        if 'mu_derived' not in solution['fit_params'].keys():
            return None
        else:
            return solution['fit_params']['mu_derived']



    def plot_posteriors(self):
        if not self.is_retrieval:
            raise Exception('HDF5 was not generated from retrieval, no posteriors found')
        
        ranges = self.compute_ranges()

        figs = []

        for solution_idx, solution_val in self.solution_iter():

            # print(solution_idx)

            mu_derived = self.get_mu_parameters(solution_val)

            tracedata = solution_val['tracedata']
            weights = solution_val['weights']

            figure_past = None

            if solution_idx > 0:
                figure_past = figs[solution_idx - 1]

            latex_names = self.fittingLatex

            if mu_derived is not None:
                latex_names.append('$\mu$ (derived)')
                tracedata = np.column_stack((tracedata, mu_derived['trace']))


            color_idx = np.float(solution_idx)/self.num_solutions

            # print('color: {}'.format(color_idx))
            ### https://matplotlib.org/users/customizing.html
            plt.rc('xtick', labelsize=10) #size of individual labels
            plt.rc('ytick', labelsize=10)
            plt.rc('axes.formatter', limits=( -4, 5 )) #scientific notation..


            fig =  corner.corner(tracedata,
                                    weights=weights,
                                    labels=latex_names,
                                    label_kwargs=dict(fontsize=20),
                                    smooth=True,
                                    scale_hist=True,
                                    quantiles=[0.16, 0.5, 0.84],
                                    show_titles=True,
                                    title_kwargs=dict(fontsize=12),
                                    range=ranges,
                                    #quantiles=[0.16, 0.5],
                                    ret=True,
                                    fill_contours=True,
                                    color=self.cmap(float(color_idx)),
                                    top_ticks=False,
                                    bins=30,
                                    fig = figure_past)
            if self.title:
                fig.gca().annotate(self.title, xy=(0.5, 1.0), xycoords="figure fraction",
                    xytext=(0, -5), textcoords="offset points",
                    ha="center", va="top", fontsize=14)

            figs.append(fig)

        plt.savefig(os.path.join(self.out_folder, '%s_posteriors.pdf' % (self.prefix)))
        self.posterior_figure_handles = figs
        self.posterior_figure_ranges  = ranges
        plt.close()

    @property
    def modelType(self):
        return self.fd['ModelParameters']['model_type'][()]

    def count_contributions(self,spectra):
        pass


    

    def plot_fitted_spectrum(self, resolution=None):

        # fitted model
        fig = plt.figure(figsize=(10.6, 7.0))
        #ax = fig.add_subplot(111)

        

        obs_spectrum = self.fd['Observed']['spectrum'][...]
        error = self.fd['Observed']['errorbars'][...]
        wlgrid = self.fd['Observed']['wlgrid'][...]
        bin_widths = self.fd['Observed']['binwidths'][...]        
        
        plt.errorbar(wlgrid,obs_spectrum, error, lw=1, color='black', alpha=0.4, ls='none', zorder=0, label='Observed')

        N = self.num_solutions
        for solution_idx, solution_val in self.solution_iter():
            if N > 1:
                label = 'Fitted model (%i)' % (solution_idx)
            else:
                label = 'Fitted model'

            try:
                binned_grid = solution_val['Spectra']['binned_wlgrid'][...]
            except KeyError:
                binned_grid = solution_val['Spectra']['bin_wlgrid'][...]
            
            native_grid = solution_val['Spectra']['native_wngrid'][...]


            plt.scatter(wlgrid, obs_spectrum, marker='d',zorder=1,**{'s': 10, 'edgecolors': 'grey','c' : self.cmap(float(solution_idx)/N) })

            self._generic_plot(binned_grid,native_grid,solution_val['Spectra'],resolution=resolution,color=self.cmap(float(solution_idx)/N),label=label)


        plt.xlim(np.min(wlgrid)-0.05*np.min(wlgrid), np.max(wlgrid)+0.05*np.max(wlgrid))
        # plt.ylim(0.0,0.006)
        plt.xlabel(r'Wavelength ($\mu$m)')
        plt.ylabel(self.modelAxis[self.modelType])

        if np.max(wlgrid) - np.min(wlgrid) > 5:
            plt.xscale('log')
            plt.tick_params(axis='x', which='minor')
            #ax.xaxis.set_minor_formatter(mpl.ticker.FormatStrFormatter("%i"))
            #ax.xaxis.set_major_formatter(mpl.ticker.FormatStrFormatter("%i"))
        plt.legend(loc='best', ncol=2, frameon=False, prop={'size':11})
        if self.title:
            plt.title(self.title, fontsize=14)
        plt.tight_layout()
        plt.savefig(os.path.join(self.out_folder, '%s_spectrum.pdf'  % (self.prefix)))
        plt.close()



    def plot_forward_spectrum(self,resolution=None):
        fig = plt.figure(figsize=(5.3, 3.5))

        spectra_out = self.forward_output()['Spectra']

        native_grid = spectra_out['native_wngrid'][...]

        try:
            wlgrid = spectra_out['binned_wlgrid'][...]
        except KeyError:
            wlgrid = spectra_out['native_wlgrid'][...]
        
    
        self._generic_plot(wlgrid,native_grid,spectra_out,resolution=resolution,alpha=1)
        plt.xlim(np.min(wlgrid)-0.05*np.min(wlgrid), np.max(wlgrid)+0.05*np.max(wlgrid))
        # plt.ylim(0.0,0.006)
        plt.xlabel(r'Wavelength ($\mu$m)')
        plt.ylabel(self.modelAxis[self.modelType])

        if np.max(wlgrid) - np.min(wlgrid) > 5:
            plt.xscale('log')
            plt.tick_params(axis='x', which='minor')
            #ax.xaxis.set_minor_formatter(mpl.ticker.FormatStrFormatter("%i"))
            #ax.xaxis.set_major_formatter(mpl.ticker.FormatStrFormatter("%i"))
        plt.legend(loc='best', ncol=2, frameon=False, prop={'size':11})
        if self.title:
            plt.title(self.title, fontsize=14)
        plt.tight_layout()
        plt.savefig(os.path.join(self.out_folder, '%s_forward_spectrum.pdf'  % (self.prefix)))
        plt.close()

    def plot_fitted_contrib(self,full=False,resolution=None):
        # fitted model

        N = self.num_solutions
        for solution_idx, solution_val in self.solution_iter():

            fig=plt.figure(figsize=(5.3*2, 3.5*2))
            ax = fig.add_subplot(111)

            

            obs_spectrum = self.fd['Observed']['spectrum'][:]
            error = self.fd['Observed']['errorbars'][...]
            wlgrid = self.fd['Observed']['wlgrid'][...]
            plot_wlgrid = wlgrid
            bin_widths = self.fd['Observed']['binwidths'][...]       
            
            plt.errorbar(wlgrid,obs_spectrum, error, lw=1, color='black', alpha=0.4, ls='none', zorder=0, label='Observed')
            self._plot_contrib(solution_val,wlgrid,ax,full=full,resolution=resolution)


            #plt.tight_layout()
            plt.savefig(os.path.join(self.out_folder, '%s_spectrum_contrib_sol%i.pdf'  % (self.prefix,solution_idx)))
            plt.close()

        plt.close('all')

    def plot_forward_contrib(self,full=False,resolution=None):
        fig=plt.figure(figsize=(5.3*2, 3.5*2))
        ax = fig.add_subplot(111)


        spectra_out = self.forward_output()['Spectra']

        native_grid = spectra_out['native_wngrid'][...]

        try:
            wlgrid = spectra_out['binned_wlgrid'][...]
        except KeyError:
            wlgrid = spectra_out['native_wlgrid'][...]

        self._generic_plot(wlgrid,native_grid,spectra_out,resolution=resolution,alpha=0.5)
        self._plot_contrib(self.forward_output(),wlgrid,ax,full=full,resolution=resolution)


        #plt.tight_layout()
        plt.savefig(os.path.join(self.out_folder, '%s_spectrum_contrib_forward.pdf'  % (self.prefix)))
        plt.close()


    def _plot_contrib(self,output,wlgrid,ax,full=False,resolution=None):


        if full:
            wlgrid = self.full_contrib_plot(output['Spectra'],wlgrid,resolution=resolution)
        else:
            wlgrid = self.simple_contrib_plot(output['Spectra'],wlgrid,resolution=resolution)

        plt.xlim(np.min(wlgrid)-0.05*np.min(wlgrid), np.max(wlgrid)+0.05*np.max(wlgrid))
        # plt.ylim(0.0,0.006)
        plt.xlabel('Wavelength ($\mu$m)')
        plt.ylabel(self.modelAxis[self.modelType])

        if np.max(wlgrid) - np.min(wlgrid) > 5:
            plt.xscale('log')
            plt.tick_params(axis='x', which='minor')
            #ax.xaxis.set_minor_formatter(mpl.ticker.FormatStrFormatter("%i"))
            #ax.xaxis.set_major_formatter(mpl.ticker.FormatStrFormatter("%i"))
        #plt.legend(loc='best', ncol=2, frameon=False, prop={'size':11})
        box = ax.get_position()
        ax.set_position([box.x0, box.y0 + box.height * 0.1,
                        box.width, box.height * 0.9])
        # Put a legend below current axis
        ax.legend(loc='upper center', bbox_to_anchor=(0.5, -0.08),
                fancybox=True, shadow=True, ncol=5)
        if self.title:
            plt.title(self.title, fontsize=14)

    def full_contrib_plot(self,spectra,wlgrid,resolution=None):
        native_grid = spectra['native_wngrid'][...]
        for contrib_name,contrib_dict in spectra['Contributions'].items():

            first_name = contrib_name

            for component_name,component_value in contrib_dict.items():
                if isinstance(component_value,h5py.Dataset):
                        continue
                total_label = '{}-{}'.format(contrib_name,component_name)
                self._generic_plot(wlgrid,native_grid,component_value,resolution,label=total_label)
        return wlgrid
    def simple_contrib_plot(self,spectra,wlgrid,resolution=None):


        binner = None
        native_grid = spectra['native_wngrid'][...]


        for contrib_name,contrib_dict in spectra['Contributions'].items():
            first_name = contrib_name
            if first_name == 'Absorption':
                for component_name,component_value in contrib_dict.items():
                    if isinstance(component_value,h5py.Dataset):
                        continue
                    total_label = '{}-{}'.format(contrib_name,component_name)
                    self._generic_plot(wlgrid,native_grid,component_value,resolution,label=total_label)
            else:
                self._generic_plot(wlgrid,native_grid,contrib_dict,resolution)
      
        return wlgrid


    def _generic_plot(self,wlgrid,native_grid,spectra,resolution,color=None,error=False,alpha=1.0,label=None):

        
        binned_error = None
        if resolution is not None:
            from taurex.binning import FluxBinner
            from taurex.util.util import create_grid_res,wnwidth_to_wlwidth
            _grid = create_grid_res(resolution,wlgrid.min()*0.9,wlgrid.max()*1.1)
            bin_wlgrid = _grid[:,0]

            bin_wngrid = 10000/_grid[:,0]

            bin_sort = bin_wngrid.argsort()

            bin_wlgrid = bin_wlgrid[bin_sort]
            bin_wngrid = bin_wngrid[bin_sort]

            bin_wnwidth = wnwidth_to_wlwidth(bin_wlgrid,_grid[bin_sort,1])
            wlgrid = _grid[bin_sort,0]
            binner = FluxBinner(bin_wngrid,bin_wnwidth)
            native_spectra = spectra['native_spectrum'][...]
            binned_spectrum = binner.bindown(native_grid,native_spectra)[1]
            try:
                native_error = spectra['native_std']
            except KeyError:
                native_error = None
            if native_error is not None:
                binned_error = binner.bindown(native_grid,native_error)[1]

        else:
            try:
                binned_spectrum = spectra['binned_spectrum'][...]
            except KeyError:
                try:
                    binned_spectrum = spectra['bin_spectrum'][...]
                except KeyError:
                    binned_spectrum = spectra['native_spectrum'][...]
            try:
                binned_error = spectra['binned_std'][...]
            except KeyError:
                binned_error = None
        plt.plot(wlgrid, binned_spectrum, label=label,alpha=alpha)          
        if binned_error is not None:
            plt.fill_between(wlgrid, binned_spectrum-binned_error,
                                binned_spectrum+binned_error,
                                alpha=0.5, zorder=-2, color=color, edgecolor='none')

            # 2 sigma
            plt.fill_between(wlgrid, binned_spectrum-2*binned_error,
                                binned_spectrum+2*binned_error,
                                alpha=0.2, zorder=-3, color=color, edgecolor='none')
        

    def plot_forward_tau(self):

        forward_output =self.forward_output()

        contribution = forward_output['Spectra']['native_tau'][...]
        #contribution = self.pickle_file['solutions'][solution_idx]['contrib_func']

        pressure = forward_output['Profiles']['pressure_profile'][:]
        wavelength = forward_output['Spectra']['native_wlgrid'][:]

        self._plot_tau(contribution,pressure,wavelength)

        plt.savefig(os.path.join(self.out_folder, '%s_tau_forward.pdf' % (self.prefix)))

        plt.close()


    def plot_fitted_tau(self):
        N = self.num_solutions
        for solution_idx, solution_val in self.solution_iter():

            contribution = solution_val['Spectra']['native_tau'][...]
            #contribution = self.pickle_file['solutions'][solution_idx]['contrib_func']

            pressure = solution_val['Profiles']['pressure_profile'][:]
            wavelength = solution_val['Spectra']['native_wlgrid'][:]

            self._plot_tau(contribution,pressure,wavelength)

            plt.savefig(os.path.join(self.out_folder, '%s_tau_sol%i.pdf' % (self.prefix,solution_idx)))

            plt.close()

    def _plot_tau(self,contribution,pressure,wavelength):
        grid = plt.GridSpec(1, 4, wspace=0.4, hspace=0.3)
        fig = plt.figure('Contribution function')
        ax1 = plt.subplot(grid[0, :3])
        plt.imshow(contribution, aspect='auto')

        ### mapping of the pressure array onto the ticks:
        y_labels = np.array([pow(10, 6), pow(10, 4), pow(10, 2), pow(10, 0), pow(10, -2), pow(10, -4)])
        y_ticks = np.zeros(len(y_labels))
        for i in range(len(y_ticks)):
            y_ticks[i] = (np.abs(pressure - y_labels[i])).argmin()  ## To find the corresponding index
        plt.yticks(y_ticks, ['$10^{%.f}$' % y for y in np.log10(y_labels) - 5])

        ### mapping of the wavelength array onto the ticks:
        x_label0 = np.ceil(np.min(wavelength) * 10) / 10.
        x_label5 = np.round(np.max(wavelength) * 10) / 10.
        x_label1 = np.round(
            pow(10, (np.log10(x_label5) - np.log10(x_label0)) * 1 / 5. + np.log10(x_label0)) * 10) / 10.0
        x_label2 = np.round(
            pow(10, (np.log10(x_label5) - np.log10(x_label0)) * 2 / 5. + np.log10(x_label0)) * 10) / 10.0
        x_label3 = np.round(
            pow(10, (np.log10(x_label5) - np.log10(x_label0)) * 3 / 5. + np.log10(x_label0)) * 10) / 10.
        x_label4 = np.round(
            pow(10, (np.log10(x_label5) - np.log10(x_label0)) * 4 / 5. + np.log10(x_label0)) * 10) / 10.

        x_labels = np.array([x_label0, x_label1, x_label2, x_label3, x_label4, x_label5])
        x_ticks = np.zeros(len(x_labels))
        for i in range(len(x_ticks)):
            x_ticks[i] = (np.abs(wavelength - x_labels[i])).argmin()  ## To find the corresponding index
        plt.xticks(x_ticks, x_labels)
        plt.gca().invert_yaxis()
        plt.gca().invert_xaxis()
        plt.xlabel("Wavelength ($\mu m$)")
        plt.ylabel("Pressure (Bar)")

        ax2 = plt.subplot(grid[0, 3])

        contribution_collapsed = np.average(contribution, axis=1)
        # contribution_collapsed = np.amax(contribution_hr, axis=1) ## good for emission
        contribution_sum = np.zeros(len(contribution_collapsed))
        for i in range(len(contribution_collapsed) - 1):
            contribution_sum[i + 1] = contribution_sum[i] + contribution_collapsed[i + 1]
        plt.plot(contribution_collapsed, pressure * pow(10, -5))

        plt.yscale('log')
        plt.gca().invert_yaxis()
        plt.gca().yaxis.tick_right()
        plt.xlabel("Contribution")


    @property
    def fittingNames(self):
        from taurex.util.util import decode_string_array
        if not self.is_retrieval:
            raise Exception('HDF5 was not generated from retrieval, no fitting names found')
        return decode_string_array(self.fd['Optimizer']['fit_parameter_names'])
    
    @property
    def fittingLatex(self):
        from taurex.util.util import decode_string_array
        if not self.is_retrieval:
            raise Exception('HDF5 was not generated from retrieval, no fitting latex found')
        return decode_string_array(self.fd['Optimizer']['fit_parameter_latex'])

    @property
    def fittingBoundaryLow(self):
        if not self.is_retrieval:
            raise Exception('HDF5 was not generated from retrieval, no fitting boundary found')
        return self.fd['Optimizer']['fit_boundary_low'][:]

    @property
    def fittingBoundaryHigh(self):
        if not self.is_retrieval:
            raise Exception('HDF5 was not generated from retrieval, no fitting boundary found')
        return self.fd['Optimizer']['fit_boundary_high'][:]


    @property
    def is_retrieval(self):
        try:
            self.fd['Output']
            self.fd['Optimizer']
            self.fd['Output']['Solutions']
            return True
        except KeyError:
            return False        

    @property
    def is_lightcurve(self):
        try:
            self.fd['Lightcurve']
            return True
        except KeyError:
            return False



def main():
    import argparse
    parser = argparse.ArgumentParser(description='Taurex-Plotter')
    parser.add_argument("-i", "--input",dest='input_file',type=str,required=True,help="Input hdf5 file from taurex")
    parser.add_argument("-P","--plot-posteriors",dest="posterior",default=False,help="Plot fitting posteriors",action='store_true')
    parser.add_argument("-x","--plot-xprofile",dest="xprofile",default=False,help="Plot molecular profiles",action='store_true')
    parser.add_argument("-t","--plot-tpprofile",dest="tpprofile",default=False,help="Plot Temperature profiles",action='store_true')
    parser.add_argument("-d","--plot-tau",dest="tau",default=False,help="Plot optical depth contribution",action="store_true")
    parser.add_argument("-s","--plot-spectrum",dest="spectrum",default=False,help="Plot spectrum",action='store_true')
    parser.add_argument("-c","--plot-contrib",dest="contrib",default=False,help="Plot contrib",action='store_true')
    parser.add_argument("-C","--full-contrib",dest="full_contrib",default=False,help="Plot detailed contribs",action="store_true")
    parser.add_argument("-a","--all",dest="all",default=False,help="Plot everythiong",action='store_true')
    parser.add_argument("-T","--title",dest="title",type=str,help="Title of plots")
    parser.add_argument("-o","--output-dir",dest="output_dir",type=str,required=True,help="output directory to store plots")
    parser.add_argument("-p","--prefix",dest="prefix",type=str,help="File prefix for outputs")
    parser.add_argument("-m","--color-map",dest="cmap",type=str,default="Paired",help="Matplotlib colormap to use")
    parser.add_argument("-R","--resolution",dest="resolution",type=float,default=None,help="Resolution to bin spectra to")
    args=parser.parse_args()

    plot_xprofile = args.xprofile or args.all
    plot_tp_profile = args.tpprofile or args.all
    plot_spectrum = args.spectrum or args.all
    plot_contrib = args.contrib or args.all
    plot_fullcontrib = args.full_contrib or args.all
    plot_posteriors = args.posterior or args.all
    plot_tau = args.tau or args.all

    plot=Plotter(args.input_file,cmap=args.cmap,
                    title=args.title,prefix=args.prefix,out_folder=args.output_dir)
    
    if plot_posteriors:
        if plot.is_retrieval:
            plot.plot_posteriors()

    if plot_xprofile:
        if plot.is_retrieval:
            plot.plot_fit_xprofile()
        else:
            plot.plot_forward_xprofile()
    if plot_spectrum:
        if plot.is_retrieval:
            plot.plot_fitted_spectrum(resolution=args.resolution)
        else:
            plot.plot_forward_spectrum(resolution=args.resolution)
    if plot_tp_profile:
        if plot.is_retrieval:
            plot.plot_fitted_tp()
        else:
            plot.plot_forward_tp()

    if plot_contrib:
        if plot.is_retrieval:
            plot.plot_fitted_contrib(full=plot_fullcontrib,resolution=args.resolution)
        else:
            plot.plot_forward_contrib(full=plot_fullcontrib,resolution=args.resolution)

    if plot_tau:
        if plot.is_retrieval:
            plot.plot_fitted_tau()
        else:
            plot.plot_forward_tau()
    
if __name__ == "__main__":
    main()




